
import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: unity_catalog_prompt_messages.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
  from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
  from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#unity_catalog_prompt_messages.proto\x12\x13mlflow.unitycatalog\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\"\x95\x02\n\x06Prompt\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x36\n\x12\x63reation_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x31\n\x07\x61liases\x18\x06 \x03(\x0b\x32 .mlflow.unitycatalog.PromptAlias\x12,\n\x04tags\x18\x07 \x03(\x0b\x32\x1e.mlflow.unitycatalog.PromptTagJ\x04\x08\x05\x10\x06R\rexperiment_id\"\xb1\x02\n\rPromptVersion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x36\n\x12\x63reation_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x10\n\x08template\x18\x06 \x01(\t\x12\x31\n\x07\x61liases\x18\x07 \x03(\x0b\x32 .mlflow.unitycatalog.PromptAlias\x12\x33\n\x04tags\x18\x08 \x03(\x0b\x32%.mlflow.unitycatalog.PromptVersionTag\"\'\n\tPromptTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\".\n\x10PromptVersionTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"-\n\x0bPromptAlias\x12\r\n\x05\x61lias\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"?\n\x12UnityCatalogSchema\x12\x14\n\x0c\x63\x61talog_name\x18\x04 \x01(\t\x12\x13\n\x0bschema_name\x18\x05 \x01(\t\"P\n\x13\x43reatePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x06prompt\x18\x02 \x01(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\"P\n\x13UpdatePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x06prompt\x18\x02 \x01(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\"#\n\x13\x44\x65letePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x16\n\x14\x44\x65letePromptResponse\" \n\x10GetPromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xb3\x01\n\x14SearchPromptsRequest\x12\x0e\n\x06\x66ilter\x18\x01 \x01(\t\x12\x41\n\x0e\x63\x61talog_schema\x18\x03 \x01(\x0b\x32\'.mlflow.unitycatalog.UnityCatalogSchemaH\x00\x12\x13\n\x0bmax_results\x18\x04 \x01(\x03\x12\x12\n\npage_token\x18\x05 \x01(\tB\n\n\x08locationJ\x04\x08\x02\x10\x03R\rexperiment_id\"^\n\x15SearchPromptsResponse\x12,\n\x07prompts\x18\x01 \x03(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"u\n\x1a\x43reatePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12:\n\x0eprompt_version\x18\x03 \x01(\x0b\x32\".mlflow.unitycatalog.PromptVersionJ\x04\x08\x02\x10\x03R\x07version\"w\n\x1aUpdatePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12:\n\x0eprompt_version\x18\x03 \x01(\x0b\x32\".mlflow.unitycatalog.PromptVersion\";\n\x1a\x44\x65letePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"8\n\x17GetPromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"T\n\x1bSearchPromptVersionsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bmax_results\x18\x02 \x01(\x03\x12\x12\n\npage_token\x18\x03 \x01(\t\"t\n\x1cSearchPromptVersionsResponse\x12;\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32\".mlflow.unitycatalog.PromptVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"E\n\x15SetPromptAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\"7\n\x18\x44\x65letePromptAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"=\n\x1eGetPromptVersionByAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"?\n\x13SetPromptTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"3\n\x16\x44\x65letePromptTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\"W\n\x1aSetPromptVersionTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\t\"K\n\x1d\x44\x65letePromptVersionTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t\"\xbe\x01\n!LinkPromptVersionsToModelsRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x11\n\tmodel_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"\xb7\x01\n\x1aLinkPromptsToTracesRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x11\n\ttrace_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"\xba\x01\n\x1fLinkPromptVersionsToRunsRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x0f\n\x07run_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"C\n\x16PromptVersionLinkEntry\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x42\x34\n\'com.databricks.api.proto.managedcatalog\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'unity_catalog_prompt_messages_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n\'com.databricks.api.proto.managedcatalog\220\001\001\240\001\001\342?\002\020\001'
    _globals['_LINKPROMPTVERSIONSTOMODELSREQUEST']._loaded_options = None
    _globals['_LINKPROMPTVERSIONSTOMODELSREQUEST']._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _globals['_LINKPROMPTSTOTRACESREQUEST']._loaded_options = None
    _globals['_LINKPROMPTSTOTRACESREQUEST']._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _globals['_LINKPROMPTVERSIONSTORUNSREQUEST']._loaded_options = None
    _globals['_LINKPROMPTVERSIONSTORUNSREQUEST']._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _globals['_PROMPTVERSIONLINKENTRY'].fields_by_name['name']._loaded_options = None
    _globals['_PROMPTVERSIONLINKENTRY'].fields_by_name['name']._serialized_options = b'\370\206\031\001'
    _globals['_PROMPTVERSIONLINKENTRY'].fields_by_name['version']._loaded_options = None
    _globals['_PROMPTVERSIONLINKENTRY'].fields_by_name['version']._serialized_options = b'\370\206\031\001'
    _globals['_PROMPT']._serialized_start=194
    _globals['_PROMPT']._serialized_end=471
    _globals['_PROMPTVERSION']._serialized_start=474
    _globals['_PROMPTVERSION']._serialized_end=779
    _globals['_PROMPTTAG']._serialized_start=781
    _globals['_PROMPTTAG']._serialized_end=820
    _globals['_PROMPTVERSIONTAG']._serialized_start=822
    _globals['_PROMPTVERSIONTAG']._serialized_end=868
    _globals['_PROMPTALIAS']._serialized_start=870
    _globals['_PROMPTALIAS']._serialized_end=915
    _globals['_UNITYCATALOGSCHEMA']._serialized_start=917
    _globals['_UNITYCATALOGSCHEMA']._serialized_end=980
    _globals['_CREATEPROMPTREQUEST']._serialized_start=982
    _globals['_CREATEPROMPTREQUEST']._serialized_end=1062
    _globals['_UPDATEPROMPTREQUEST']._serialized_start=1064
    _globals['_UPDATEPROMPTREQUEST']._serialized_end=1144
    _globals['_DELETEPROMPTREQUEST']._serialized_start=1146
    _globals['_DELETEPROMPTREQUEST']._serialized_end=1181
    _globals['_DELETEPROMPTRESPONSE']._serialized_start=1183
    _globals['_DELETEPROMPTRESPONSE']._serialized_end=1205
    _globals['_GETPROMPTREQUEST']._serialized_start=1207
    _globals['_GETPROMPTREQUEST']._serialized_end=1239
    _globals['_SEARCHPROMPTSREQUEST']._serialized_start=1242
    _globals['_SEARCHPROMPTSREQUEST']._serialized_end=1421
    _globals['_SEARCHPROMPTSRESPONSE']._serialized_start=1423
    _globals['_SEARCHPROMPTSRESPONSE']._serialized_end=1517
    _globals['_CREATEPROMPTVERSIONREQUEST']._serialized_start=1519
    _globals['_CREATEPROMPTVERSIONREQUEST']._serialized_end=1636
    _globals['_UPDATEPROMPTVERSIONREQUEST']._serialized_start=1638
    _globals['_UPDATEPROMPTVERSIONREQUEST']._serialized_end=1757
    _globals['_DELETEPROMPTVERSIONREQUEST']._serialized_start=1759
    _globals['_DELETEPROMPTVERSIONREQUEST']._serialized_end=1818
    _globals['_GETPROMPTVERSIONREQUEST']._serialized_start=1820
    _globals['_GETPROMPTVERSIONREQUEST']._serialized_end=1876
    _globals['_SEARCHPROMPTVERSIONSREQUEST']._serialized_start=1878
    _globals['_SEARCHPROMPTVERSIONSREQUEST']._serialized_end=1962
    _globals['_SEARCHPROMPTVERSIONSRESPONSE']._serialized_start=1964
    _globals['_SEARCHPROMPTVERSIONSRESPONSE']._serialized_end=2080
    _globals['_SETPROMPTALIASREQUEST']._serialized_start=2082
    _globals['_SETPROMPTALIASREQUEST']._serialized_end=2151
    _globals['_DELETEPROMPTALIASREQUEST']._serialized_start=2153
    _globals['_DELETEPROMPTALIASREQUEST']._serialized_end=2208
    _globals['_GETPROMPTVERSIONBYALIASREQUEST']._serialized_start=2210
    _globals['_GETPROMPTVERSIONBYALIASREQUEST']._serialized_end=2271
    _globals['_SETPROMPTTAGREQUEST']._serialized_start=2273
    _globals['_SETPROMPTTAGREQUEST']._serialized_end=2336
    _globals['_DELETEPROMPTTAGREQUEST']._serialized_start=2338
    _globals['_DELETEPROMPTTAGREQUEST']._serialized_end=2389
    _globals['_SETPROMPTVERSIONTAGREQUEST']._serialized_start=2391
    _globals['_SETPROMPTVERSIONTAGREQUEST']._serialized_end=2478
    _globals['_DELETEPROMPTVERSIONTAGREQUEST']._serialized_start=2480
    _globals['_DELETEPROMPTVERSIONTAGREQUEST']._serialized_end=2555
    _globals['_LINKPROMPTVERSIONSTOMODELSREQUEST']._serialized_start=2558
    _globals['_LINKPROMPTVERSIONSTOMODELSREQUEST']._serialized_end=2748
    _globals['_LINKPROMPTSTOTRACESREQUEST']._serialized_start=2751
    _globals['_LINKPROMPTSTOTRACESREQUEST']._serialized_end=2934
    _globals['_LINKPROMPTVERSIONSTORUNSREQUEST']._serialized_start=2937
    _globals['_LINKPROMPTVERSIONSTORUNSREQUEST']._serialized_end=3123
    _globals['_PROMPTVERSIONLINKENTRY']._serialized_start=3125
    _globals['_PROMPTVERSIONLINKENTRY']._serialized_end=3192
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: unity_catalog_prompt_messages.proto
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
  from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
  from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
  from . import databricks_pb2 as databricks__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#unity_catalog_prompt_messages.proto\x12\x13mlflow.unitycatalog\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a\x10\x64\x61tabricks.proto\x1a\x15scalapb/scalapb.proto\"\x95\x02\n\x06Prompt\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x36\n\x12\x63reation_timestamp\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x64\x65scription\x18\x04 \x01(\t\x12\x31\n\x07\x61liases\x18\x06 \x03(\x0b\x32 .mlflow.unitycatalog.PromptAlias\x12,\n\x04tags\x18\x07 \x03(\x0b\x32\x1e.mlflow.unitycatalog.PromptTagJ\x04\x08\x05\x10\x06R\rexperiment_id\"\xb1\x02\n\rPromptVersion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x36\n\x12\x63reation_timestamp\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12:\n\x16last_updated_timestamp\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x10\n\x08template\x18\x06 \x01(\t\x12\x31\n\x07\x61liases\x18\x07 \x03(\x0b\x32 .mlflow.unitycatalog.PromptAlias\x12\x33\n\x04tags\x18\x08 \x03(\x0b\x32%.mlflow.unitycatalog.PromptVersionTag\"\'\n\tPromptTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\".\n\x10PromptVersionTag\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"-\n\x0bPromptAlias\x12\r\n\x05\x61lias\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"?\n\x12UnityCatalogSchema\x12\x14\n\x0c\x63\x61talog_name\x18\x04 \x01(\t\x12\x13\n\x0bschema_name\x18\x05 \x01(\t\"P\n\x13\x43reatePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x06prompt\x18\x02 \x01(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\"P\n\x13UpdatePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12+\n\x06prompt\x18\x02 \x01(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\"#\n\x13\x44\x65letePromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x16\n\x14\x44\x65letePromptResponse\" \n\x10GetPromptRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xb3\x01\n\x14SearchPromptsRequest\x12\x0e\n\x06\x66ilter\x18\x01 \x01(\t\x12\x41\n\x0e\x63\x61talog_schema\x18\x03 \x01(\x0b\x32\'.mlflow.unitycatalog.UnityCatalogSchemaH\x00\x12\x13\n\x0bmax_results\x18\x04 \x01(\x03\x12\x12\n\npage_token\x18\x05 \x01(\tB\n\n\x08locationJ\x04\x08\x02\x10\x03R\rexperiment_id\"^\n\x15SearchPromptsResponse\x12,\n\x07prompts\x18\x01 \x03(\x0b\x32\x1b.mlflow.unitycatalog.Prompt\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"u\n\x1a\x43reatePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12:\n\x0eprompt_version\x18\x03 \x01(\x0b\x32\".mlflow.unitycatalog.PromptVersionJ\x04\x08\x02\x10\x03R\x07version\"w\n\x1aUpdatePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12:\n\x0eprompt_version\x18\x03 \x01(\x0b\x32\".mlflow.unitycatalog.PromptVersion\";\n\x1a\x44\x65letePromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"8\n\x17GetPromptVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\"T\n\x1bSearchPromptVersionsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bmax_results\x18\x02 \x01(\x03\x12\x12\n\npage_token\x18\x03 \x01(\t\"t\n\x1cSearchPromptVersionsResponse\x12;\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32\".mlflow.unitycatalog.PromptVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"E\n\x15SetPromptAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\"7\n\x18\x44\x65letePromptAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"=\n\x1eGetPromptVersionByAliasRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x61lias\x18\x02 \x01(\t\"?\n\x13SetPromptTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"3\n\x16\x44\x65letePromptTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\"W\n\x1aSetPromptVersionTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t\x12\r\n\x05value\x18\x04 \x01(\t\"K\n\x1d\x44\x65letePromptVersionTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0b\n\x03key\x18\x03 \x01(\t\"\xbe\x01\n!LinkPromptVersionsToModelsRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x11\n\tmodel_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"\xb7\x01\n\x1aLinkPromptsToTracesRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x11\n\ttrace_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"\xba\x01\n\x1fLinkPromptVersionsToRunsRequest\x12\x44\n\x0fprompt_versions\x18\x01 \x03(\x0b\x32+.mlflow.unitycatalog.PromptVersionLinkEntry\x12\x0f\n\x07run_ids\x18\x02 \x03(\t:@\xe2?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]\"C\n\x16PromptVersionLinkEntry\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\xf8\x86\x19\x01\x12\x15\n\x07version\x18\x02 \x01(\tB\x04\xf8\x86\x19\x01\x42\x34\n\'com.databricks.api.proto.managedcatalog\x90\x01\x01\xa0\x01\x01\xe2?\x02\x10\x01')



  _PROMPT = DESCRIPTOR.message_types_by_name['Prompt']
  _PROMPTVERSION = DESCRIPTOR.message_types_by_name['PromptVersion']
  _PROMPTTAG = DESCRIPTOR.message_types_by_name['PromptTag']
  _PROMPTVERSIONTAG = DESCRIPTOR.message_types_by_name['PromptVersionTag']
  _PROMPTALIAS = DESCRIPTOR.message_types_by_name['PromptAlias']
  _UNITYCATALOGSCHEMA = DESCRIPTOR.message_types_by_name['UnityCatalogSchema']
  _CREATEPROMPTREQUEST = DESCRIPTOR.message_types_by_name['CreatePromptRequest']
  _UPDATEPROMPTREQUEST = DESCRIPTOR.message_types_by_name['UpdatePromptRequest']
  _DELETEPROMPTREQUEST = DESCRIPTOR.message_types_by_name['DeletePromptRequest']
  _DELETEPROMPTRESPONSE = DESCRIPTOR.message_types_by_name['DeletePromptResponse']
  _GETPROMPTREQUEST = DESCRIPTOR.message_types_by_name['GetPromptRequest']
  _SEARCHPROMPTSREQUEST = DESCRIPTOR.message_types_by_name['SearchPromptsRequest']
  _SEARCHPROMPTSRESPONSE = DESCRIPTOR.message_types_by_name['SearchPromptsResponse']
  _CREATEPROMPTVERSIONREQUEST = DESCRIPTOR.message_types_by_name['CreatePromptVersionRequest']
  _UPDATEPROMPTVERSIONREQUEST = DESCRIPTOR.message_types_by_name['UpdatePromptVersionRequest']
  _DELETEPROMPTVERSIONREQUEST = DESCRIPTOR.message_types_by_name['DeletePromptVersionRequest']
  _GETPROMPTVERSIONREQUEST = DESCRIPTOR.message_types_by_name['GetPromptVersionRequest']
  _SEARCHPROMPTVERSIONSREQUEST = DESCRIPTOR.message_types_by_name['SearchPromptVersionsRequest']
  _SEARCHPROMPTVERSIONSRESPONSE = DESCRIPTOR.message_types_by_name['SearchPromptVersionsResponse']
  _SETPROMPTALIASREQUEST = DESCRIPTOR.message_types_by_name['SetPromptAliasRequest']
  _DELETEPROMPTALIASREQUEST = DESCRIPTOR.message_types_by_name['DeletePromptAliasRequest']
  _GETPROMPTVERSIONBYALIASREQUEST = DESCRIPTOR.message_types_by_name['GetPromptVersionByAliasRequest']
  _SETPROMPTTAGREQUEST = DESCRIPTOR.message_types_by_name['SetPromptTagRequest']
  _DELETEPROMPTTAGREQUEST = DESCRIPTOR.message_types_by_name['DeletePromptTagRequest']
  _SETPROMPTVERSIONTAGREQUEST = DESCRIPTOR.message_types_by_name['SetPromptVersionTagRequest']
  _DELETEPROMPTVERSIONTAGREQUEST = DESCRIPTOR.message_types_by_name['DeletePromptVersionTagRequest']
  _LINKPROMPTVERSIONSTOMODELSREQUEST = DESCRIPTOR.message_types_by_name['LinkPromptVersionsToModelsRequest']
  _LINKPROMPTSTOTRACESREQUEST = DESCRIPTOR.message_types_by_name['LinkPromptsToTracesRequest']
  _LINKPROMPTVERSIONSTORUNSREQUEST = DESCRIPTOR.message_types_by_name['LinkPromptVersionsToRunsRequest']
  _PROMPTVERSIONLINKENTRY = DESCRIPTOR.message_types_by_name['PromptVersionLinkEntry']
  Prompt = _reflection.GeneratedProtocolMessageType('Prompt', (_message.Message,), {
    'DESCRIPTOR' : _PROMPT,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.Prompt)
    })
  _sym_db.RegisterMessage(Prompt)

  PromptVersion = _reflection.GeneratedProtocolMessageType('PromptVersion', (_message.Message,), {
    'DESCRIPTOR' : _PROMPTVERSION,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.PromptVersion)
    })
  _sym_db.RegisterMessage(PromptVersion)

  PromptTag = _reflection.GeneratedProtocolMessageType('PromptTag', (_message.Message,), {
    'DESCRIPTOR' : _PROMPTTAG,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.PromptTag)
    })
  _sym_db.RegisterMessage(PromptTag)

  PromptVersionTag = _reflection.GeneratedProtocolMessageType('PromptVersionTag', (_message.Message,), {
    'DESCRIPTOR' : _PROMPTVERSIONTAG,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.PromptVersionTag)
    })
  _sym_db.RegisterMessage(PromptVersionTag)

  PromptAlias = _reflection.GeneratedProtocolMessageType('PromptAlias', (_message.Message,), {
    'DESCRIPTOR' : _PROMPTALIAS,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.PromptAlias)
    })
  _sym_db.RegisterMessage(PromptAlias)

  UnityCatalogSchema = _reflection.GeneratedProtocolMessageType('UnityCatalogSchema', (_message.Message,), {
    'DESCRIPTOR' : _UNITYCATALOGSCHEMA,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.UnityCatalogSchema)
    })
  _sym_db.RegisterMessage(UnityCatalogSchema)

  CreatePromptRequest = _reflection.GeneratedProtocolMessageType('CreatePromptRequest', (_message.Message,), {
    'DESCRIPTOR' : _CREATEPROMPTREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.CreatePromptRequest)
    })
  _sym_db.RegisterMessage(CreatePromptRequest)

  UpdatePromptRequest = _reflection.GeneratedProtocolMessageType('UpdatePromptRequest', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEPROMPTREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.UpdatePromptRequest)
    })
  _sym_db.RegisterMessage(UpdatePromptRequest)

  DeletePromptRequest = _reflection.GeneratedProtocolMessageType('DeletePromptRequest', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptRequest)
    })
  _sym_db.RegisterMessage(DeletePromptRequest)

  DeletePromptResponse = _reflection.GeneratedProtocolMessageType('DeletePromptResponse', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTRESPONSE,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptResponse)
    })
  _sym_db.RegisterMessage(DeletePromptResponse)

  GetPromptRequest = _reflection.GeneratedProtocolMessageType('GetPromptRequest', (_message.Message,), {
    'DESCRIPTOR' : _GETPROMPTREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.GetPromptRequest)
    })
  _sym_db.RegisterMessage(GetPromptRequest)

  SearchPromptsRequest = _reflection.GeneratedProtocolMessageType('SearchPromptsRequest', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHPROMPTSREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SearchPromptsRequest)
    })
  _sym_db.RegisterMessage(SearchPromptsRequest)

  SearchPromptsResponse = _reflection.GeneratedProtocolMessageType('SearchPromptsResponse', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHPROMPTSRESPONSE,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SearchPromptsResponse)
    })
  _sym_db.RegisterMessage(SearchPromptsResponse)

  CreatePromptVersionRequest = _reflection.GeneratedProtocolMessageType('CreatePromptVersionRequest', (_message.Message,), {
    'DESCRIPTOR' : _CREATEPROMPTVERSIONREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.CreatePromptVersionRequest)
    })
  _sym_db.RegisterMessage(CreatePromptVersionRequest)

  UpdatePromptVersionRequest = _reflection.GeneratedProtocolMessageType('UpdatePromptVersionRequest', (_message.Message,), {
    'DESCRIPTOR' : _UPDATEPROMPTVERSIONREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.UpdatePromptVersionRequest)
    })
  _sym_db.RegisterMessage(UpdatePromptVersionRequest)

  DeletePromptVersionRequest = _reflection.GeneratedProtocolMessageType('DeletePromptVersionRequest', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTVERSIONREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptVersionRequest)
    })
  _sym_db.RegisterMessage(DeletePromptVersionRequest)

  GetPromptVersionRequest = _reflection.GeneratedProtocolMessageType('GetPromptVersionRequest', (_message.Message,), {
    'DESCRIPTOR' : _GETPROMPTVERSIONREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.GetPromptVersionRequest)
    })
  _sym_db.RegisterMessage(GetPromptVersionRequest)

  SearchPromptVersionsRequest = _reflection.GeneratedProtocolMessageType('SearchPromptVersionsRequest', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHPROMPTVERSIONSREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SearchPromptVersionsRequest)
    })
  _sym_db.RegisterMessage(SearchPromptVersionsRequest)

  SearchPromptVersionsResponse = _reflection.GeneratedProtocolMessageType('SearchPromptVersionsResponse', (_message.Message,), {
    'DESCRIPTOR' : _SEARCHPROMPTVERSIONSRESPONSE,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SearchPromptVersionsResponse)
    })
  _sym_db.RegisterMessage(SearchPromptVersionsResponse)

  SetPromptAliasRequest = _reflection.GeneratedProtocolMessageType('SetPromptAliasRequest', (_message.Message,), {
    'DESCRIPTOR' : _SETPROMPTALIASREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SetPromptAliasRequest)
    })
  _sym_db.RegisterMessage(SetPromptAliasRequest)

  DeletePromptAliasRequest = _reflection.GeneratedProtocolMessageType('DeletePromptAliasRequest', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTALIASREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptAliasRequest)
    })
  _sym_db.RegisterMessage(DeletePromptAliasRequest)

  GetPromptVersionByAliasRequest = _reflection.GeneratedProtocolMessageType('GetPromptVersionByAliasRequest', (_message.Message,), {
    'DESCRIPTOR' : _GETPROMPTVERSIONBYALIASREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.GetPromptVersionByAliasRequest)
    })
  _sym_db.RegisterMessage(GetPromptVersionByAliasRequest)

  SetPromptTagRequest = _reflection.GeneratedProtocolMessageType('SetPromptTagRequest', (_message.Message,), {
    'DESCRIPTOR' : _SETPROMPTTAGREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SetPromptTagRequest)
    })
  _sym_db.RegisterMessage(SetPromptTagRequest)

  DeletePromptTagRequest = _reflection.GeneratedProtocolMessageType('DeletePromptTagRequest', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTTAGREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptTagRequest)
    })
  _sym_db.RegisterMessage(DeletePromptTagRequest)

  SetPromptVersionTagRequest = _reflection.GeneratedProtocolMessageType('SetPromptVersionTagRequest', (_message.Message,), {
    'DESCRIPTOR' : _SETPROMPTVERSIONTAGREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.SetPromptVersionTagRequest)
    })
  _sym_db.RegisterMessage(SetPromptVersionTagRequest)

  DeletePromptVersionTagRequest = _reflection.GeneratedProtocolMessageType('DeletePromptVersionTagRequest', (_message.Message,), {
    'DESCRIPTOR' : _DELETEPROMPTVERSIONTAGREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.DeletePromptVersionTagRequest)
    })
  _sym_db.RegisterMessage(DeletePromptVersionTagRequest)

  LinkPromptVersionsToModelsRequest = _reflection.GeneratedProtocolMessageType('LinkPromptVersionsToModelsRequest', (_message.Message,), {
    'DESCRIPTOR' : _LINKPROMPTVERSIONSTOMODELSREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.LinkPromptVersionsToModelsRequest)
    })
  _sym_db.RegisterMessage(LinkPromptVersionsToModelsRequest)

  LinkPromptsToTracesRequest = _reflection.GeneratedProtocolMessageType('LinkPromptsToTracesRequest', (_message.Message,), {
    'DESCRIPTOR' : _LINKPROMPTSTOTRACESREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.LinkPromptsToTracesRequest)
    })
  _sym_db.RegisterMessage(LinkPromptsToTracesRequest)

  LinkPromptVersionsToRunsRequest = _reflection.GeneratedProtocolMessageType('LinkPromptVersionsToRunsRequest', (_message.Message,), {
    'DESCRIPTOR' : _LINKPROMPTVERSIONSTORUNSREQUEST,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.LinkPromptVersionsToRunsRequest)
    })
  _sym_db.RegisterMessage(LinkPromptVersionsToRunsRequest)

  PromptVersionLinkEntry = _reflection.GeneratedProtocolMessageType('PromptVersionLinkEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROMPTVERSIONLINKENTRY,
    '__module__' : 'unity_catalog_prompt_messages_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.unitycatalog.PromptVersionLinkEntry)
    })
  _sym_db.RegisterMessage(PromptVersionLinkEntry)

  if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n\'com.databricks.api.proto.managedcatalog\220\001\001\240\001\001\342?\002\020\001'
    _LINKPROMPTVERSIONSTOMODELSREQUEST._options = None
    _LINKPROMPTVERSIONSTOMODELSREQUEST._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _LINKPROMPTSTOTRACESREQUEST._options = None
    _LINKPROMPTSTOTRACESREQUEST._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _LINKPROMPTVERSIONSTORUNSREQUEST._options = None
    _LINKPROMPTVERSIONSTORUNSREQUEST._serialized_options = b'\342?=\n;UcModelRegistryRoutableRPC[com.google.protobuf.empty.Empty]'
    _PROMPTVERSIONLINKENTRY.fields_by_name['name']._options = None
    _PROMPTVERSIONLINKENTRY.fields_by_name['name']._serialized_options = b'\370\206\031\001'
    _PROMPTVERSIONLINKENTRY.fields_by_name['version']._options = None
    _PROMPTVERSIONLINKENTRY.fields_by_name['version']._serialized_options = b'\370\206\031\001'
    _PROMPT._serialized_start=194
    _PROMPT._serialized_end=471
    _PROMPTVERSION._serialized_start=474
    _PROMPTVERSION._serialized_end=779
    _PROMPTTAG._serialized_start=781
    _PROMPTTAG._serialized_end=820
    _PROMPTVERSIONTAG._serialized_start=822
    _PROMPTVERSIONTAG._serialized_end=868
    _PROMPTALIAS._serialized_start=870
    _PROMPTALIAS._serialized_end=915
    _UNITYCATALOGSCHEMA._serialized_start=917
    _UNITYCATALOGSCHEMA._serialized_end=980
    _CREATEPROMPTREQUEST._serialized_start=982
    _CREATEPROMPTREQUEST._serialized_end=1062
    _UPDATEPROMPTREQUEST._serialized_start=1064
    _UPDATEPROMPTREQUEST._serialized_end=1144
    _DELETEPROMPTREQUEST._serialized_start=1146
    _DELETEPROMPTREQUEST._serialized_end=1181
    _DELETEPROMPTRESPONSE._serialized_start=1183
    _DELETEPROMPTRESPONSE._serialized_end=1205
    _GETPROMPTREQUEST._serialized_start=1207
    _GETPROMPTREQUEST._serialized_end=1239
    _SEARCHPROMPTSREQUEST._serialized_start=1242
    _SEARCHPROMPTSREQUEST._serialized_end=1421
    _SEARCHPROMPTSRESPONSE._serialized_start=1423
    _SEARCHPROMPTSRESPONSE._serialized_end=1517
    _CREATEPROMPTVERSIONREQUEST._serialized_start=1519
    _CREATEPROMPTVERSIONREQUEST._serialized_end=1636
    _UPDATEPROMPTVERSIONREQUEST._serialized_start=1638
    _UPDATEPROMPTVERSIONREQUEST._serialized_end=1757
    _DELETEPROMPTVERSIONREQUEST._serialized_start=1759
    _DELETEPROMPTVERSIONREQUEST._serialized_end=1818
    _GETPROMPTVERSIONREQUEST._serialized_start=1820
    _GETPROMPTVERSIONREQUEST._serialized_end=1876
    _SEARCHPROMPTVERSIONSREQUEST._serialized_start=1878
    _SEARCHPROMPTVERSIONSREQUEST._serialized_end=1962
    _SEARCHPROMPTVERSIONSRESPONSE._serialized_start=1964
    _SEARCHPROMPTVERSIONSRESPONSE._serialized_end=2080
    _SETPROMPTALIASREQUEST._serialized_start=2082
    _SETPROMPTALIASREQUEST._serialized_end=2151
    _DELETEPROMPTALIASREQUEST._serialized_start=2153
    _DELETEPROMPTALIASREQUEST._serialized_end=2208
    _GETPROMPTVERSIONBYALIASREQUEST._serialized_start=2210
    _GETPROMPTVERSIONBYALIASREQUEST._serialized_end=2271
    _SETPROMPTTAGREQUEST._serialized_start=2273
    _SETPROMPTTAGREQUEST._serialized_end=2336
    _DELETEPROMPTTAGREQUEST._serialized_start=2338
    _DELETEPROMPTTAGREQUEST._serialized_end=2389
    _SETPROMPTVERSIONTAGREQUEST._serialized_start=2391
    _SETPROMPTVERSIONTAGREQUEST._serialized_end=2478
    _DELETEPROMPTVERSIONTAGREQUEST._serialized_start=2480
    _DELETEPROMPTVERSIONTAGREQUEST._serialized_end=2555
    _LINKPROMPTVERSIONSTOMODELSREQUEST._serialized_start=2558
    _LINKPROMPTVERSIONSTOMODELSREQUEST._serialized_end=2748
    _LINKPROMPTSTOTRACESREQUEST._serialized_start=2751
    _LINKPROMPTSTOTRACESREQUEST._serialized_end=2934
    _LINKPROMPTVERSIONSTORUNSREQUEST._serialized_start=2937
    _LINKPROMPTVERSIONSTORUNSREQUEST._serialized_end=3123
    _PROMPTVERSIONLINKENTRY._serialized_start=3125
    _PROMPTVERSIONLINKENTRY._serialized_end=3192
  # @@protoc_insertion_point(module_scope)




from mlflow.entities.model_registry._model_registry_entity import _ModelRegistryEntity
from mlflow.protos.model_registry_pb2 import RegisteredModelAlias as ProtoRegisteredModelAlias


class RegisteredModelAlias(_ModelRegistryEntity):
    """Alias object associated with a registered model."""

    def __init__(self, alias, version):
        self._alias = alias
        self._version = version

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def alias(self):
        """String name of the alias."""
        return self._alias

    @property
    def version(self):
        """String model version number that the alias points to."""
        return self._version

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.alias, proto.version)

    def to_proto(self):
        alias_proto = ProtoRegisteredModelAlias()
        alias_proto.alias = self.alias
        alias_proto.version = self.version
        return alias_proto



from mlflow.entities._mlflow_object import _MlflowObject
from mlflow.protos.service_pb2 import ModelInput as ProtoModelInput


class LoggedModelInput(_MlflowObject):
    """ModelInput object associated with a Run."""

    def __init__(self, model_id: str):
        self._model_id = model_id

    def __eq__(self, other: _MlflowObject) -> bool:
        if type(other) is type(self):
            return self.__dict__ == other.__dict__
        return False

    @property
    def model_id(self) -> str:
        """Model ID."""
        return self._model_id

    def to_proto(self):
        return ProtoModelInput(model_id=self._model_id)

    @classmethod
    def from_proto(cls, proto):
        return cls(proto.model_id)



"""
Utilities for validating user inputs such as metric names and parameter names.
"""

import json
import logging
import numbers
import posixpath
import re
from typing import Optional

from mlflow.entities import Dataset, DatasetInput, InputTag, Param, RunTag
from mlflow.entities.model_registry.prompt_version import PROMPT_TEXT_TAG_KEY
from mlflow.environment_variables import (
    MLFLOW_ARTIFACT_LOCATION_MAX_LENGTH,
    MLFLOW_TRUNCATE_LONG_VALUES,
)
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.utils.os import is_windows
from mlflow.utils.string_utils import is_string_type

_logger = logging.getLogger(__name__)

# Regex for valid run IDs: must be an alphanumeric string of length 1 to 256.
_RUN_ID_REGEX = re.compile(r"^[a-zA-Z0-9][\w\-]{0,255}$")

# Regex: starting with an alphanumeric, optionally followed by up to 63 characters
# including alphanumerics, underscores, or dashes.
_EXPERIMENT_ID_REGEX = re.compile(r"^[a-zA-Z0-9][\w\-]{0,63}$")

# Regex for valid registered model alias names: may only contain alphanumerics,
# underscores, and dashes.
_REGISTERED_MODEL_ALIAS_REGEX = re.compile(r"^[\w\-]*$")

# Regex for valid registered model alias to prevent conflict with version aliases.
_REGISTERED_MODEL_ALIAS_VERSION_REGEX = re.compile(r"^[vV]\d+$")

_BAD_ALIAS_CHARACTERS_MESSAGE = (
    "Names may only contain alphanumerics, underscores (_), and dashes (-)."
)

_MISSING_KEY_NAME_MESSAGE = "A key name must be provided."

MAX_PARAMS_TAGS_PER_BATCH = 100
MAX_METRICS_PER_BATCH = 1000
MAX_DATASETS_PER_BATCH = 1000
MAX_ENTITIES_PER_BATCH = 1000
MAX_BATCH_LOG_REQUEST_SIZE = int(1e6)
MAX_PARAM_VAL_LENGTH = 6000
MAX_TAG_VAL_LENGTH = 8000
MAX_EXPERIMENT_NAME_LENGTH = 500
MAX_EXPERIMENT_TAG_KEY_LENGTH = 250
MAX_EXPERIMENT_TAG_VAL_LENGTH = 5000
MAX_ENTITY_KEY_LENGTH = 250
MAX_MODEL_REGISTRY_TAG_KEY_LENGTH = 250
MAX_MODEL_REGISTRY_TAG_VALUE_LENGTH = 100_000
MAX_EXPERIMENTS_LISTED_PER_PAGE = 50000
MAX_DATASET_NAME_SIZE = 500
MAX_DATASET_DIGEST_SIZE = 36
# 1MB -1, the db limit for MEDIUMTEXT column is 16MB, but
# we restrict to 1MB because user might log lots of datasets
# to a single run, 16MB increases burden on db
MAX_DATASET_SCHEMA_SIZE = 1048575
MAX_DATASET_SOURCE_SIZE = 65535  # 64KB -1 (the db limit for TEXT column)
MAX_DATASET_PROFILE_SIZE = 16777215  # 16MB -1 (the db limit for MEDIUMTEXT column)
MAX_INPUT_TAG_KEY_SIZE = 255
MAX_INPUT_TAG_VALUE_SIZE = 500
MAX_REGISTERED_MODEL_ALIAS_LENGTH = 255
MAX_TRACE_TAG_KEY_LENGTH = 250
MAX_TRACE_TAG_VAL_LENGTH = 8000

PARAM_VALIDATION_MSG = """

The cause of this error is typically due to repeated calls
to an individual run_id event logging.

Incorrect Example:
---------------------------------------
with mlflow.start_run():
    mlflow.log_param("depth", 3)
    mlflow.log_param("depth", 5)
---------------------------------------

Which will throw an MlflowException for overwriting a
logged parameter.

Correct Example:
---------------------------------------
with mlflow.start_run():
    with mlflow.start_run(nested=True):
        mlflow.log_param("depth", 3)
    with mlflow.start_run(nested=True):
        mlflow.log_param("depth", 5)
---------------------------------------

Which will create a new nested run for each individual
model and prevent parameter key collisions within the
tracking store."""


def invalid_value(path, value, message=None):
    """
    Compose a standardized error message for invalid parameter values.
    """
    formattedValue = json.dumps(value, sort_keys=True, separators=(",", ":"))

    if message:
        return f"Invalid value {formattedValue} for parameter '{path}' supplied: {message}"
    else:
        return f"Invalid value {formattedValue} for parameter '{path}' supplied."


def missing_value(path):
    return f"Missing value for required parameter '{path}'."


def not_integer_value(path, value):
    return f"Parameter '{path}' must be an integer, got '{value}'."


def exceeds_maximum_length(path, limit):
    return f"'{path}' exceeds the maximum length of {limit} characters"


def append_to_json_path(currentPath, value):
    if not currentPath:
        return value

    if value.startswith("["):
        return f"{currentPath}{value}"

    return f"{currentPath}.{value}"


def bad_path_message(name):
    return (
        "Names may be treated as files in certain cases, and must not resolve to other names"
        f" when treated as such. This name would resolve to {posixpath.normpath(name)!r}"
    )


def validate_param_and_metric_name(name):
    # In windows system valid param and metric names: may only contain slashes, alphanumerics,
    # underscores, periods, dashes, and spaces.
    if is_windows():
        return re.match(r"^[/\w.\- ]*$", name)

    # For other system valid param and metric names: may only contain slashes, alphanumerics,
    # underscores, periods, dashes, colons, and spaces.
    return re.match(r"^[/\w.\- :]*$", name)


def bad_character_message():
    # Valid param and metric names may only contain slashes, alphanumerics, underscores,
    # periods, dashes, colons, and spaces. For windows param and metric names can not contain colon
    msg = (
        "Names may only contain alphanumerics, underscores (_), dashes (-), periods (.),"
        " spaces ( ){} and slashes (/)."
    )
    return msg.format("") if is_windows() else msg.format(", colon(:)")


def path_not_unique(name):
    norm = posixpath.normpath(name)
    return norm != name or norm == "." or norm.startswith("..") or norm.startswith("/")


def _validate_metric_name(name, path="name"):
    """Check that `name` is a valid metric name and raise an exception if it isn't."""
    if name is None:
        raise MlflowException(
            invalid_value(path, name, f"Metric name cannot be None. {_MISSING_KEY_NAME_MESSAGE}"),
            error_code=INVALID_PARAMETER_VALUE,
        )
    if not validate_param_and_metric_name(name):
        raise MlflowException(
            invalid_value(path, name, bad_character_message()),
            INVALID_PARAMETER_VALUE,
        )
    if path_not_unique(name):
        raise MlflowException(
            invalid_value(path, name, bad_path_message(name)),
            INVALID_PARAMETER_VALUE,
        )


def _is_numeric(value):
    """
    Returns True if the passed-in value is numeric.
    """
    # Note that `isinstance(bool_value, numbers.Number)` returns `True` because `bool` is a
    # subclass of `int`.
    return not isinstance(value, bool) and isinstance(value, numbers.Number)


def _validate_metric(key, value, timestamp, step, path=""):
    """
    Check that a metric with the specified key, value, timestamp, and step is valid and raise an
    exception if it isn't.
    """
    _validate_metric_name(key, append_to_json_path(path, "name"))

    # If invocated via log_metric, no prior validation of the presence of the value was done.
    if value is None:
        raise MlflowException(
            missing_value(append_to_json_path(path, "value")),
            INVALID_PARAMETER_VALUE,
        )

    # value must be a Number
    # since bool is an instance of Number check for bool additionally
    if not _is_numeric(value):
        raise MlflowException(
            invalid_value(
                append_to_json_path(path, "value"),
                value,
                f"(timestamp={timestamp}). "
                f"Please specify value as a valid double (64-bit floating point)",
            ),
            INVALID_PARAMETER_VALUE,
        )

    if not isinstance(timestamp, numbers.Number) or timestamp < 0:
        raise MlflowException(
            invalid_value(
                append_to_json_path(path, "timestamp"),
                timestamp,
                f"metric '{key}' (value={value}). "
                f"Timestamp must be a nonnegative long (64-bit integer) ",
            ),
            INVALID_PARAMETER_VALUE,
        )

    if not isinstance(step, numbers.Number):
        raise MlflowException(
            invalid_value(
                append_to_json_path(path, "step"),
                step,
                f"metric '{key}' (value={value}). Step must be a valid long (64-bit integer).",
            ),
            INVALID_PARAMETER_VALUE,
        )

    _validate_length_limit("Metric name", MAX_ENTITY_KEY_LENGTH, key)


def _validate_param(key, value, path=""):
    """
    Check that a param with the specified key & value is valid and raise an exception if it
    isn't.
    """
    _validate_param_name(key, append_to_json_path(path, "key"))
    return Param(
        _validate_length_limit("Param key", MAX_ENTITY_KEY_LENGTH, key),
        _validate_length_limit("Param value", MAX_PARAM_VAL_LENGTH, value, truncate=True),
    )


def _validate_tag(key, value, path=""):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key, append_to_json_path(path, "key"))
    return RunTag(
        _validate_length_limit(append_to_json_path(path, "key"), MAX_ENTITY_KEY_LENGTH, key),
        _validate_length_limit(
            append_to_json_path(path, "value"), MAX_TAG_VAL_LENGTH, value, truncate=True
        ),
    )


def _validate_experiment_tag(key, value):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key)
    _validate_length_limit("key", MAX_EXPERIMENT_TAG_KEY_LENGTH, key)
    _validate_length_limit("value", MAX_EXPERIMENT_TAG_VAL_LENGTH, value)


def _validate_registered_model_tag(key, value):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key)
    _validate_length_limit("key", MAX_MODEL_REGISTRY_TAG_KEY_LENGTH, key)
    _validate_length_limit("value", MAX_MODEL_REGISTRY_TAG_VALUE_LENGTH, value)


def _validate_model_version_tag(key, value):
    """
    Check that a tag with the specified key & value is valid and raise an exception if it isn't.
    """
    _validate_tag_name(key)
    _validate_tag_value(value)
    _validate_length_limit("key", MAX_MODEL_REGISTRY_TAG_KEY_LENGTH, key)

    # Check prompt text tag particularly for showing friendly error message
    if key == PROMPT_TEXT_TAG_KEY and len(value) > MAX_MODEL_REGISTRY_TAG_VALUE_LENGTH:
        raise MlflowException.invalid_parameter_value(
            f"Prompt text exceeds max length of {MAX_MODEL_REGISTRY_TAG_VALUE_LENGTH} characters.",
        )

    _validate_length_limit("value", MAX_MODEL_REGISTRY_TAG_VALUE_LENGTH, value)


def _validate_param_keys_unique(params):
    """Ensures that duplicate param keys are not present in the `log_batch()` params argument"""
    unique_keys = []
    dupe_keys = []
    for param in params:
        if param.key not in unique_keys:
            unique_keys.append(param.key)
        else:
            dupe_keys.append(param.key)

    if dupe_keys:
        raise MlflowException(
            f"Duplicate parameter keys have been submitted: {dupe_keys}. Please ensure "
            "the request contains only one param value per param key.",
            INVALID_PARAMETER_VALUE,
        )


def _validate_param_name(name, path="key"):
    """Check that `name` is a valid parameter name and raise an exception if it isn't."""
    if name is None:
        raise MlflowException(
            invalid_value(path, "", _MISSING_KEY_NAME_MESSAGE),
            error_code=INVALID_PARAMETER_VALUE,
        )
    if not validate_param_and_metric_name(name):
        raise MlflowException(
            invalid_value(path, name, bad_character_message()),
            INVALID_PARAMETER_VALUE,
        )
    if path_not_unique(name):
        raise MlflowException(
            invalid_value(path, name, bad_path_message(name)),
            INVALID_PARAMETER_VALUE,
        )


def _validate_tag_name(name, path="key"):
    """Check that `name` is a valid tag name and raise an exception if it isn't."""
    # Reuse param & metric check.
    if name is None:
        raise MlflowException(
            missing_value(path),
            error_code=INVALID_PARAMETER_VALUE,
        )
    if not validate_param_and_metric_name(name):
        raise MlflowException(
            invalid_value(path, name, bad_character_message()),
            INVALID_PARAMETER_VALUE,
        )
    if path_not_unique(name):
        raise MlflowException(
            invalid_value(path, name, bad_path_message(name)),
            INVALID_PARAMETER_VALUE,
        )


def _validate_length_limit(entity_name, limit, value, *, truncate=False):
    if value is None:
        return None

    if len(value) <= limit:
        return value

    if truncate and MLFLOW_TRUNCATE_LONG_VALUES.get():
        _logger.warning(
            f"{entity_name} '{value[:100]}...' ({len(value)} characters) is truncated to "
            f"{limit} characters to meet the length limit."
        )
        return value[:limit]

    raise MlflowException(
        exceeds_maximum_length(entity_name, limit),
        error_code=INVALID_PARAMETER_VALUE,
    )


def _validate_run_id(run_id, path="run_id"):
    """Check that `run_id` is a valid run ID and raise an exception if it isn't."""
    if _RUN_ID_REGEX.match(run_id) is None:
        raise MlflowException(invalid_value(path, run_id), error_code=INVALID_PARAMETER_VALUE)


def _validate_experiment_id(exp_id):
    """Check that `experiment_id`is a valid string or None, raise an exception if it isn't."""
    if exp_id is not None and _EXPERIMENT_ID_REGEX.match(exp_id) is None:
        raise MlflowException(
            f"Invalid experiment ID: '{exp_id}'", error_code=INVALID_PARAMETER_VALUE
        )


def _validate_batch_limit(entity_name, limit, length):
    if length > limit:
        error_msg = (
            f"A batch logging request can contain at most {limit} {entity_name}. "
            f"Got {length} {entity_name}. Please split up {entity_name} across multiple"
            " requests and try again."
        )
        raise MlflowException(error_msg, error_code=INVALID_PARAMETER_VALUE)


def _validate_batch_log_limits(metrics, params, tags):
    """Validate that the provided batched logging arguments are within expected limits."""
    _validate_batch_limit(entity_name="metrics", limit=MAX_METRICS_PER_BATCH, length=len(metrics))
    _validate_batch_limit(entity_name="params", limit=MAX_PARAMS_TAGS_PER_BATCH, length=len(params))
    _validate_batch_limit(entity_name="tags", limit=MAX_PARAMS_TAGS_PER_BATCH, length=len(tags))
    total_length = len(metrics) + len(params) + len(tags)
    _validate_batch_limit(
        entity_name="metrics, params, and tags",
        limit=MAX_ENTITIES_PER_BATCH,
        length=total_length,
    )


def _validate_batch_log_data(metrics, params, tags):
    for index, metric in enumerate(metrics):
        path = f"metrics[{index}]"
        _validate_metric(metric.key, metric.value, metric.timestamp, metric.step, path=path)
    return (
        metrics,
        [_validate_param(p.key, p.value, path=f"params[{idx}]") for (idx, p) in enumerate(params)],
        [_validate_tag(t.key, t.value, path=f"tags[{idx}]") for (idx, t) in enumerate(tags)],
    )


def _validate_batch_log_api_req(json_req):
    if len(json_req) > MAX_BATCH_LOG_REQUEST_SIZE:
        error_msg = (
            "Batched logging API requests must be at most {limit} bytes, got a "
            "request of size {size}."
        ).format(limit=MAX_BATCH_LOG_REQUEST_SIZE, size=len(json_req))
        raise MlflowException(error_msg, error_code=INVALID_PARAMETER_VALUE)


def _validate_experiment_name(experiment_name):
    """Check that `experiment_name` is a valid string and raise an exception if it isn't."""
    if experiment_name == "" or experiment_name is None:
        raise MlflowException(
            f"Invalid experiment name: '{experiment_name}'",
            error_code=INVALID_PARAMETER_VALUE,
        )

    if not is_string_type(experiment_name):
        raise MlflowException(
            f"Invalid experiment name: {experiment_name}. Expects a string.",
            error_code=INVALID_PARAMETER_VALUE,
        )

    if len(experiment_name) > MAX_EXPERIMENT_NAME_LENGTH:
        raise MlflowException.invalid_parameter_value(
            exceeds_maximum_length("name", MAX_EXPERIMENT_NAME_LENGTH)
        )


def _validate_experiment_id_type(experiment_id):
    """
    Check that a user-provided experiment_id is either a string, int, or None and raise an
    exception if it isn't.
    """
    if experiment_id is not None and not isinstance(experiment_id, (str, int)):
        raise MlflowException(
            f"Invalid experiment id: {experiment_id} of type {type(experiment_id)}. "
            "Must be one of str, int, or None.",
            error_code=INVALID_PARAMETER_VALUE,
        )


def _validate_model_name(model_name):
    if model_name is None or model_name == "":
        raise MlflowException(missing_value("name"), error_code=INVALID_PARAMETER_VALUE)


def _validate_model_renaming(model_new_name):
    if model_new_name is None or model_new_name == "":
        raise MlflowException(missing_value("new_name"), error_code=INVALID_PARAMETER_VALUE)


def _validate_model_version(model_version):
    try:
        model_version = int(model_version)
    except ValueError:
        raise MlflowException(
            not_integer_value("version", model_version), error_code=INVALID_PARAMETER_VALUE
        )


def _validate_model_alias_name(model_alias_name):
    if model_alias_name is None or model_alias_name == "":
        raise MlflowException(
            "Registered model alias name cannot be empty.", INVALID_PARAMETER_VALUE
        )
    if not _REGISTERED_MODEL_ALIAS_REGEX.match(model_alias_name):
        raise MlflowException(
            f"Invalid alias name: '{model_alias_name}'. {_BAD_ALIAS_CHARACTERS_MESSAGE}",
            INVALID_PARAMETER_VALUE,
        )
    _validate_length_limit(
        "Registered model alias name",
        MAX_REGISTERED_MODEL_ALIAS_LENGTH,
        model_alias_name,
    )
    if model_alias_name.lower() == "latest":
        raise MlflowException(
            "'latest' alias name (case insensitive) is reserved.",
            INVALID_PARAMETER_VALUE,
        )
    if _REGISTERED_MODEL_ALIAS_VERSION_REGEX.match(model_alias_name):
        raise MlflowException(
            f"Version alias name '{model_alias_name}' is reserved.",
            INVALID_PARAMETER_VALUE,
        )


def _validate_experiment_artifact_location(artifact_location):
    if artifact_location is not None and artifact_location.startswith("runs:"):
        raise MlflowException(
            f"Artifact location cannot be a runs:/ URI. Given: '{artifact_location}'",
            error_code=INVALID_PARAMETER_VALUE,
        )


def _validate_db_type_string(db_type):
    """validates db_type parsed from DB URI is supported"""
    from mlflow.store.db.db_types import DATABASE_ENGINES

    if db_type not in DATABASE_ENGINES:
        error_msg = (
            f"Invalid database engine: '{db_type}'. "
            f"Supported database engines are {', '.join(DATABASE_ENGINES)}"
        )
        raise MlflowException(error_msg, INVALID_PARAMETER_VALUE)


def _validate_model_version_or_stage_exists(version, stage):
    if version and stage:
        raise MlflowException("version and stage cannot be set together", INVALID_PARAMETER_VALUE)

    if not (version or stage):
        raise MlflowException("version or stage must be set", INVALID_PARAMETER_VALUE)


def _validate_tag_value(value):
    if value is None:
        raise MlflowException("Tag value cannot be None", INVALID_PARAMETER_VALUE)


def _validate_dataset_inputs(dataset_inputs: list[DatasetInput]):
    for dataset_input in dataset_inputs:
        _validate_dataset(dataset_input.dataset)
        _validate_input_tags(dataset_input.tags)


def _validate_dataset(dataset: Dataset):
    if dataset is None:
        raise MlflowException("Dataset cannot be None", INVALID_PARAMETER_VALUE)
    if dataset.name is None:
        raise MlflowException("Dataset name cannot be None", INVALID_PARAMETER_VALUE)
    if dataset.digest is None:
        raise MlflowException("Dataset digest cannot be None", INVALID_PARAMETER_VALUE)
    if dataset.source_type is None:
        raise MlflowException("Dataset source_type cannot be None", INVALID_PARAMETER_VALUE)
    if dataset.source is None:
        raise MlflowException("Dataset source cannot be None", INVALID_PARAMETER_VALUE)
    if len(dataset.name) > MAX_DATASET_NAME_SIZE:
        raise MlflowException(
            exceeds_maximum_length("name", MAX_DATASET_NAME_SIZE),
            INVALID_PARAMETER_VALUE,
        )
    if len(dataset.digest) > MAX_DATASET_DIGEST_SIZE:
        raise MlflowException(
            exceeds_maximum_length("digest", MAX_DATASET_DIGEST_SIZE),
            INVALID_PARAMETER_VALUE,
        )
    if len(dataset.source) > MAX_DATASET_SOURCE_SIZE:
        raise MlflowException(
            exceeds_maximum_length("source", MAX_DATASET_SOURCE_SIZE),
            INVALID_PARAMETER_VALUE,
        )
    if dataset.schema is not None and len(dataset.schema) > MAX_DATASET_SCHEMA_SIZE:
        raise MlflowException(
            exceeds_maximum_length("schema", MAX_DATASET_SCHEMA_SIZE),
            INVALID_PARAMETER_VALUE,
        )
    if dataset.profile is not None and len(dataset.profile) > MAX_DATASET_PROFILE_SIZE:
        raise MlflowException(
            exceeds_maximum_length("profile", MAX_DATASET_PROFILE_SIZE),
            INVALID_PARAMETER_VALUE,
        )


def _validate_input_tags(input_tags: list[InputTag]):
    for input_tag in input_tags:
        _validate_input_tag(input_tag)


def _validate_input_tag(input_tag: InputTag):
    if input_tag is None:
        raise MlflowException("InputTag cannot be None", INVALID_PARAMETER_VALUE)
    if input_tag.key is None:
        raise MlflowException("InputTag key cannot be None", INVALID_PARAMETER_VALUE)
    if input_tag.value is None:
        raise MlflowException("InputTag value cannot be None", INVALID_PARAMETER_VALUE)
    if len(input_tag.key) > MAX_INPUT_TAG_KEY_SIZE:
        raise MlflowException(
            exceeds_maximum_length("key", MAX_INPUT_TAG_KEY_SIZE),
            INVALID_PARAMETER_VALUE,
        )
    if len(input_tag.value) > MAX_INPUT_TAG_VALUE_SIZE:
        raise MlflowException(
            exceeds_maximum_length("value", MAX_INPUT_TAG_VALUE_SIZE),
            INVALID_PARAMETER_VALUE,
        )


def _validate_username(username):
    if username is None or username == "":
        raise MlflowException("Username cannot be empty.", INVALID_PARAMETER_VALUE)


def _validate_password(password) -> None:
    if password is None or len(password) < 12:
        raise MlflowException.invalid_parameter_value(
            "Password must be a string longer than 12 characters."
        )


def _validate_trace_tag(key, value):
    _validate_tag_name(key)
    key = _validate_length_limit("key", MAX_TRACE_TAG_KEY_LENGTH, key)
    value = _validate_length_limit("value", MAX_TRACE_TAG_VAL_LENGTH, value, truncate=True)
    return key, value


def _validate_experiment_artifact_location_length(artifact_location: str):
    max_length = MLFLOW_ARTIFACT_LOCATION_MAX_LENGTH.get()
    if len(artifact_location) > max_length:
        raise MlflowException(
            "Invalid artifact path length. The length of the artifact path cannot be "
            f"greater than {max_length} characters. To configure this limit, please set the "
            "MLFLOW_ARTIFACT_LOCATION_MAX_LENGTH environment variable.",
            INVALID_PARAMETER_VALUE,
        )


def _validate_logged_model_name(name: Optional[str]) -> None:
    if name is None:
        return

    bad_chars = ("/", ":", ".", "%", '"', "'")
    if not name or any(c in name for c in bad_chars):
        raise MlflowException(
            f"Invalid model name ({name!r}) provided. Model name must be a non-empty string "
            f"and cannot contain the following characters: {bad_chars}",
            INVALID_PARAMETER_VALUE,
        )



import graphene


class ErrorDetail(graphene.ObjectType):
    # NOTE: This is not an exhaustive list, might need to add more things in the future if needed.
    field = graphene.String()
    message = graphene.String()


class ApiError(graphene.ObjectType):
    code = graphene.String()
    message = graphene.String()
    help_url = graphene.String()
    trace_id = graphene.String()
    error_details = graphene.List(ErrorDetail)



import logging
import math
from collections import namedtuple
from contextlib import contextmanager
from typing import Optional

import numpy as np
import pandas as pd
from sklearn import metrics as sk_metrics

import mlflow
from mlflow import MlflowException
from mlflow.environment_variables import _MLFLOW_EVALUATE_SUPPRESS_CLASSIFICATION_ERRORS
from mlflow.models.evaluation.artifacts import CsvEvaluationArtifact
from mlflow.models.evaluation.base import EvaluationMetric, EvaluationResult, _ModelType
from mlflow.models.evaluation.default_evaluator import (
    BuiltInEvaluator,
    _extract_raw_model,
    _get_aggregate_metrics_values,
)
from mlflow.models.utils import plot_lines

_logger = logging.getLogger(__name__)


_Curve = namedtuple("_Curve", ["plot_fn", "plot_fn_args", "auc"])


class ClassifierEvaluator(BuiltInEvaluator):
    """
    A built-in evaluator for classifier models.
    """

    name = "classifier"

    @classmethod
    def can_evaluate(cls, *, model_type, evaluator_config, **kwargs):
        # TODO: Also the model needs to be pyfunc model, not function or endpoint URI
        return model_type == _ModelType.CLASSIFIER

    def _evaluate(
        self,
        model: Optional["mlflow.pyfunc.PyFuncModel"],
        extra_metrics: list[EvaluationMetric],
        custom_artifacts=None,
        **kwargs,
    ) -> Optional[EvaluationResult]:
        # Get classification config
        self.y_true = self.dataset.labels_data
        self.label_list = self.evaluator_config.get("label_list")
        self.pos_label = self.evaluator_config.get("pos_label")
        self.sample_weights = self.evaluator_config.get("sample_weights")
        if self.pos_label and self.label_list and self.pos_label not in self.label_list:
            raise MlflowException.invalid_parameter_value(
                f"'pos_label' {self.pos_label} must exist in 'label_list' {self.label_list}."
            )

        # Check if the model_type is consistent with ground truth labels
        inferred_model_type = _infer_model_type_by_labels(self.y_true)
        if _ModelType.CLASSIFIER != inferred_model_type:
            _logger.warning(
                f"According to the evaluation dataset label values, the model type looks like "
                f"{inferred_model_type}, but you specified model type 'classifier'. Please "
                f"verify that you set the `model_type` and `dataset` arguments correctly."
            )

        # Run model prediction
        input_df = self.X.copy_to_avoid_mutation()
        self.y_pred, self.y_probs = self._generate_model_predictions(model, input_df)

        self._validate_label_list()

        self._compute_builtin_metrics(model)
        self.evaluate_metrics(extra_metrics, prediction=self.y_pred, target=self.y_true)
        self.evaluate_and_log_custom_artifacts(
            custom_artifacts, prediction=self.y_pred, target=self.y_true
        )

        # Log metrics and artifacts
        self.log_metrics()
        self.log_eval_table(self.y_pred)

        if len(self.label_list) == 2:
            self._log_binary_classifier_artifacts()
        else:
            self._log_multiclass_classifier_artifacts()
        self._log_confusion_matrix()

        return EvaluationResult(
            metrics=self.aggregate_metrics, artifacts=self.artifacts, run_id=self.run_id
        )

    def _generate_model_predictions(self, model, input_df):
        predict_fn, predict_proba_fn = _extract_predict_fn_and_prodict_proba_fn(model)
        # Classifier model is guaranteed to output single column of predictions
        y_pred = self.dataset.predictions_data if model is None else predict_fn(input_df)

        # Predict class probabilities if the model supports it
        y_probs = predict_proba_fn(input_df) if predict_proba_fn is not None else None
        return y_pred, y_probs

    def _validate_label_list(self):
        if self.label_list is None:
            # If label list is not specified, infer label list from model output
            self.label_list = np.unique(np.concatenate([self.y_true, self.y_pred]))
        else:
            # np.where only works for numpy array, not list
            self.label_list = np.array(self.label_list)

        if len(self.label_list) < 2:
            raise MlflowException(
                "Evaluation dataset for classification must contain at least two unique "
                f"labels, but only {len(self.label_list)} unique labels were found.",
                "Please provide a 'label_list' parameter in 'evaluator_config' with all "
                "possible classes, e.g., evaluator_config={{'label_list': [0, 1]}}.",
            )

        # sort label_list ASC, for binary classification it makes sure the last one is pos label
        self.label_list.sort()

        if len(self.label_list) == 2:
            if self.pos_label is None:
                self.pos_label = self.label_list[-1]
            else:
                if self.pos_label in self.label_list:
                    self.label_list = np.delete(
                        self.label_list, np.where(self.label_list == self.pos_label)
                    )
                self.label_list = np.append(self.label_list, self.pos_label)
            with _suppress_class_imbalance_errors(IndexError, log_warning=False):
                _logger.info(
                    "The evaluation dataset is inferred as binary dataset, positive label is "
                    f"{self.label_list[1]}, negative label is {self.label_list[0]}."
                )
        else:
            _logger.info(
                "The evaluation dataset is inferred as multiclass dataset, number of classes "
                f"is inferred as {len(self.label_list)}. If this is incorrect, please specify the "
                "`label_list` parameter in `evaluator_config`."
            )

    def _compute_builtin_metrics(self, model):
        self._evaluate_sklearn_model_score_if_scorable(model, self.y_true, self.sample_weights)

        if len(self.label_list) == 2:
            metrics = _get_binary_classifier_metrics(
                y_true=self.y_true,
                y_pred=self.y_pred,
                y_proba=self.y_probs,
                labels=self.label_list,
                pos_label=self.pos_label,
                sample_weights=self.sample_weights,
            )
            if metrics:
                self.metrics_values.update(_get_aggregate_metrics_values(metrics))
                self._compute_roc_and_pr_curve()
        else:
            average = self.evaluator_config.get("average", "weighted")
            metrics = _get_multiclass_classifier_metrics(
                y_true=self.y_true,
                y_pred=self.y_pred,
                y_proba=self.y_probs,
                labels=self.label_list,
                average=average,
                sample_weights=self.sample_weights,
            )
            if metrics:
                self.metrics_values.update(_get_aggregate_metrics_values(metrics))

    def _compute_roc_and_pr_curve(self):
        if self.y_probs is not None:
            with _suppress_class_imbalance_errors(ValueError, log_warning=False):
                self.roc_curve = _gen_classifier_curve(
                    is_binomial=True,
                    y=self.y_true,
                    y_probs=self.y_probs[:, 1],
                    labels=self.label_list,
                    pos_label=self.pos_label,
                    curve_type="roc",
                    sample_weights=self.sample_weights,
                )

                self.metrics_values.update(
                    _get_aggregate_metrics_values({"roc_auc": self.roc_curve.auc})
                )
            with _suppress_class_imbalance_errors(ValueError, log_warning=False):
                self.pr_curve = _gen_classifier_curve(
                    is_binomial=True,
                    y=self.y_true,
                    y_probs=self.y_probs[:, 1],
                    labels=self.label_list,
                    pos_label=self.pos_label,
                    curve_type="pr",
                    sample_weights=self.sample_weights,
                )

                self.metrics_values.update(
                    _get_aggregate_metrics_values({"precision_recall_auc": self.pr_curve.auc})
                )

    def _log_pandas_df_artifact(self, pandas_df, artifact_name):
        artifact_file_name = f"{artifact_name}.csv"
        artifact_file_local_path = self.temp_dir.path(artifact_file_name)
        pandas_df.to_csv(artifact_file_local_path, index=False)
        mlflow.log_artifact(artifact_file_local_path)
        artifact = CsvEvaluationArtifact(
            uri=mlflow.get_artifact_uri(artifact_file_name),
            content=pandas_df,
        )
        artifact._load(artifact_file_local_path)
        self.artifacts[artifact_name] = artifact

    def _log_multiclass_classifier_artifacts(self):
        per_class_metrics_collection_df = _get_classifier_per_class_metrics_collection_df(
            y=self.y_true,
            y_pred=self.y_pred,
            labels=self.label_list,
            sample_weights=self.sample_weights,
        )

        log_roc_pr_curve = False
        if self.y_probs is not None:
            with _suppress_class_imbalance_errors(TypeError, log_warning=False):
                self._log_calibration_curve()

            max_classes_for_multiclass_roc_pr = self.evaluator_config.get(
                "max_classes_for_multiclass_roc_pr", 10
            )
            if len(self.label_list) <= max_classes_for_multiclass_roc_pr:
                log_roc_pr_curve = True
            else:
                _logger.warning(
                    f"The classifier num_classes > {max_classes_for_multiclass_roc_pr}, skip "
                    f"logging ROC curve and Precision-Recall curve. You can add evaluator config "
                    f"'max_classes_for_multiclass_roc_pr' to increase the threshold."
                )

        if log_roc_pr_curve:
            roc_curve = _gen_classifier_curve(
                is_binomial=False,
                y=self.y_true,
                y_probs=self.y_probs,
                labels=self.label_list,
                pos_label=self.pos_label,
                curve_type="roc",
                sample_weights=self.sample_weights,
            )

            def plot_roc_curve():
                roc_curve.plot_fn(**roc_curve.plot_fn_args)

            self._log_image_artifact(plot_roc_curve, "roc_curve_plot")
            per_class_metrics_collection_df["roc_auc"] = roc_curve.auc

            pr_curve = _gen_classifier_curve(
                is_binomial=False,
                y=self.y_true,
                y_probs=self.y_probs,
                labels=self.label_list,
                pos_label=self.pos_label,
                curve_type="pr",
                sample_weights=self.sample_weights,
            )

            def plot_pr_curve():
                pr_curve.plot_fn(**pr_curve.plot_fn_args)

            self._log_image_artifact(plot_pr_curve, "precision_recall_curve_plot")
            per_class_metrics_collection_df["precision_recall_auc"] = pr_curve.auc

        self._log_pandas_df_artifact(per_class_metrics_collection_df, "per_class_metrics")

    def _log_roc_curve(self):
        def _plot_roc_curve():
            self.roc_curve.plot_fn(**self.roc_curve.plot_fn_args)

        self._log_image_artifact(_plot_roc_curve, "roc_curve_plot")

    def _log_precision_recall_curve(self):
        def _plot_pr_curve():
            self.pr_curve.plot_fn(**self.pr_curve.plot_fn_args)

        self._log_image_artifact(_plot_pr_curve, "precision_recall_curve_plot")

    def _log_lift_curve(self):
        from mlflow.models.evaluation.lift_curve import plot_lift_curve

        def _plot_lift_curve():
            return plot_lift_curve(self.y_true, self.y_probs, pos_label=self.pos_label)

        self._log_image_artifact(_plot_lift_curve, "lift_curve_plot")

    def _log_calibration_curve(self):
        from mlflow.models.evaluation.calibration_curve import plot_calibration_curve

        def _plot_calibration_curve():
            return plot_calibration_curve(
                y_true=self.y_true,
                y_probs=self.y_probs,
                pos_label=self.pos_label,
                calibration_config={
                    k: v for k, v in self.evaluator_config.items() if k.startswith("calibration_")
                },
                label_list=self.label_list,
            )

        self._log_image_artifact(_plot_calibration_curve, "calibration_curve_plot")

    def _log_binary_classifier_artifacts(self):
        if self.y_probs is not None:
            with _suppress_class_imbalance_errors(log_warning=False):
                self._log_roc_curve()
            with _suppress_class_imbalance_errors(log_warning=False):
                self._log_precision_recall_curve()
            with _suppress_class_imbalance_errors(ValueError, log_warning=False):
                self._log_lift_curve()
            with _suppress_class_imbalance_errors(TypeError, log_warning=False):
                self._log_calibration_curve()

    def _log_confusion_matrix(self):
        """
        Helper method for logging confusion matrix
        """
        # normalize the confusion matrix, keep consistent with sklearn autologging.
        confusion_matrix = sk_metrics.confusion_matrix(
            self.y_true,
            self.y_pred,
            labels=self.label_list,
            normalize="true",
            sample_weight=self.sample_weights,
        )

        def plot_confusion_matrix():
            import matplotlib
            import matplotlib.pyplot as plt

            with matplotlib.rc_context(
                {
                    "font.size": min(8, math.ceil(50.0 / len(self.label_list))),
                    "axes.labelsize": 8,
                }
            ):
                _, ax = plt.subplots(1, 1, figsize=(6.0, 4.0), dpi=175)
                disp = sk_metrics.ConfusionMatrixDisplay(
                    confusion_matrix=confusion_matrix,
                    display_labels=self.label_list,
                ).plot(cmap="Blues", ax=ax)
                disp.ax_.set_title("Normalized confusion matrix")

        if hasattr(sk_metrics, "ConfusionMatrixDisplay"):
            self._log_image_artifact(
                plot_confusion_matrix,
                "confusion_matrix",
            )
        return


def _is_categorical(values):
    """
    Infer whether input values are categorical on best effort.
    Return True represent they are categorical, return False represent we cannot determine result.
    """
    dtype_name = pd.Series(values).convert_dtypes().dtype.name.lower()
    return dtype_name in ["category", "string", "boolean"]


def _is_continuous(values):
    """
    Infer whether input values is continuous on best effort.
    Return True represent they are continuous, return False represent we cannot determine result.
    """
    dtype_name = pd.Series(values).convert_dtypes().dtype.name.lower()
    return dtype_name.startswith("float")


def _infer_model_type_by_labels(labels):
    """
    Infer model type by target values.
    """
    if _is_categorical(labels):
        return _ModelType.CLASSIFIER
    elif _is_continuous(labels):
        return _ModelType.REGRESSOR
    else:
        return None  # Unknown


def _extract_predict_fn_and_prodict_proba_fn(model):
    predict_fn = None
    predict_proba_fn = None

    _, raw_model = _extract_raw_model(model)

    if raw_model is not None:
        predict_fn = raw_model.predict
        predict_proba_fn = getattr(raw_model, "predict_proba", None)
        try:
            from mlflow.xgboost import (
                _wrapped_xgboost_model_predict_fn,
                _wrapped_xgboost_model_predict_proba_fn,
            )

            # Because shap evaluation will pass evaluation data in ndarray format
            # (without feature names), if set validate_features=True it will raise error.
            predict_fn = _wrapped_xgboost_model_predict_fn(raw_model, validate_features=False)
            predict_proba_fn = _wrapped_xgboost_model_predict_proba_fn(
                raw_model, validate_features=False
            )
        except ImportError:
            pass
    elif model is not None:
        predict_fn = model.predict

    return predict_fn, predict_proba_fn


@contextmanager
def _suppress_class_imbalance_errors(exception_type=Exception, log_warning=True):
    """
    Exception handler context manager to suppress Exceptions if the private environment
    variable `_MLFLOW_EVALUATE_SUPPRESS_CLASSIFICATION_ERRORS` is set to `True`.
    The purpose of this handler is to prevent an evaluation call for a binary or multiclass
    classification automl run from aborting due to an extreme minority class imbalance
    encountered during iterative training cycles due to the non deterministic sampling
    behavior of Spark's DataFrame.sample() API.
    The Exceptions caught in the usage of this are broad and are designed purely to not
    interrupt the iterative hyperparameter tuning process. Final evaluations are done
    in a more deterministic (but expensive) fashion.
    """
    try:
        yield
    except exception_type as e:
        if _MLFLOW_EVALUATE_SUPPRESS_CLASSIFICATION_ERRORS.get():
            if log_warning:
                _logger.warning(
                    "Failed to calculate metrics due to class imbalance. "
                    "This is expected when the dataset is imbalanced."
                )
        else:
            raise e


def _get_binary_sum_up_label_pred_prob(positive_class_index, positive_class, y, y_pred, y_probs):
    y = np.array(y)
    y_bin = np.where(y == positive_class, 1, 0)
    y_pred_bin = None
    y_prob_bin = None
    if y_pred is not None:
        y_pred = np.array(y_pred)
        y_pred_bin = np.where(y_pred == positive_class, 1, 0)

    if y_probs is not None:
        y_probs = np.array(y_probs)
        y_prob_bin = y_probs[:, positive_class_index]

    return y_bin, y_pred_bin, y_prob_bin


def _get_common_classifier_metrics(
    *, y_true, y_pred, y_proba, labels, average, pos_label, sample_weights
):
    metrics = {
        "example_count": len(y_true),
        "accuracy_score": sk_metrics.accuracy_score(y_true, y_pred, sample_weight=sample_weights),
        "recall_score": sk_metrics.recall_score(
            y_true,
            y_pred,
            average=average,
            pos_label=pos_label,
            sample_weight=sample_weights,
        ),
        "precision_score": sk_metrics.precision_score(
            y_true,
            y_pred,
            average=average,
            pos_label=pos_label,
            sample_weight=sample_weights,
        ),
        "f1_score": sk_metrics.f1_score(
            y_true,
            y_pred,
            average=average,
            pos_label=pos_label,
            sample_weight=sample_weights,
        ),
    }

    if y_proba is not None:
        with _suppress_class_imbalance_errors(ValueError):
            metrics["log_loss"] = sk_metrics.log_loss(
                y_true, y_proba, labels=labels, sample_weight=sample_weights
            )
    return metrics


def _get_binary_classifier_metrics(
    *, y_true, y_pred, y_proba=None, labels=None, pos_label=1, sample_weights=None
):
    with _suppress_class_imbalance_errors(ValueError):
        tn, fp, fn, tp = sk_metrics.confusion_matrix(y_true, y_pred, labels=labels).ravel()
        return {
            "true_negatives": tn,
            "false_positives": fp,
            "false_negatives": fn,
            "true_positives": tp,
            **_get_common_classifier_metrics(
                y_true=y_true,
                y_pred=y_pred,
                y_proba=y_proba,
                labels=labels,
                average="binary",
                pos_label=pos_label,
                sample_weights=sample_weights,
            ),
        }


def _get_multiclass_classifier_metrics(
    *,
    y_true,
    y_pred,
    y_proba=None,
    labels=None,
    average="weighted",
    sample_weights=None,
):
    metrics = _get_common_classifier_metrics(
        y_true=y_true,
        y_pred=y_pred,
        y_proba=y_proba,
        labels=labels,
        average=average,
        pos_label=None,
        sample_weights=sample_weights,
    )
    if average in ("macro", "weighted") and y_proba is not None:
        metrics.update(
            roc_auc=sk_metrics.roc_auc_score(
                y_true=y_true,
                y_score=y_proba,
                sample_weight=sample_weights,
                average=average,
                multi_class="ovr",
            )
        )
    return metrics


def _get_classifier_per_class_metrics_collection_df(y, y_pred, labels, sample_weights):
    per_class_metrics_list = []
    for positive_class_index, positive_class in enumerate(labels):
        (
            y_bin,
            y_pred_bin,
            _,
        ) = _get_binary_sum_up_label_pred_prob(
            positive_class_index, positive_class, y, y_pred, None
        )
        per_class_metrics = {"positive_class": positive_class}
        binary_classifier_metrics = _get_binary_classifier_metrics(
            y_true=y_bin,
            y_pred=y_pred_bin,
            labels=[0, 1],  # Use binary labels for per-class metrics
            pos_label=1,
            sample_weights=sample_weights,
        )
        if binary_classifier_metrics:
            per_class_metrics.update(binary_classifier_metrics)
        per_class_metrics_list.append(per_class_metrics)

    return pd.DataFrame(per_class_metrics_list)


_Curve = namedtuple("_Curve", ["plot_fn", "plot_fn_args", "auc"])


def _gen_classifier_curve(
    is_binomial,
    y,
    y_probs,
    labels,
    pos_label,
    curve_type,
    sample_weights,
):
    """
    Generate precision-recall curve or ROC curve for classifier.

    Args:
        is_binomial: True if it is binary classifier otherwise False
        y: True label values
        y_probs: if binary classifier, the predicted probability for positive class.
                  if multiclass classifier, the predicted probabilities for all classes.
        labels: The set of labels.
        pos_label: The label of the positive class.
        curve_type: "pr" or "roc"
        sample_weights: Optional sample weights.

    Returns:
        An instance of "_Curve" which includes attributes "plot_fn", "plot_fn_args", "auc".
    """
    if curve_type == "roc":

        def gen_line_x_y_label_auc(_y, _y_prob, _pos_label):
            fpr, tpr, _ = sk_metrics.roc_curve(
                _y,
                _y_prob,
                sample_weight=sample_weights,
                # For multiclass classification where a one-vs-rest ROC curve is produced for each
                # class, the positive label is binarized and should not be included in the plot
                # legend
                pos_label=_pos_label if _pos_label == pos_label else None,
            )

            auc = sk_metrics.roc_auc_score(y_true=_y, y_score=_y_prob, sample_weight=sample_weights)
            return fpr, tpr, f"AUC={auc:.3f}", auc

        xlabel = "False Positive Rate"
        ylabel = "True Positive Rate"
        title = "ROC curve"
        if pos_label:
            xlabel = f"False Positive Rate (Positive label: {pos_label})"
            ylabel = f"True Positive Rate (Positive label: {pos_label})"
    elif curve_type == "pr":

        def gen_line_x_y_label_auc(_y, _y_prob, _pos_label):
            precision, recall, _ = sk_metrics.precision_recall_curve(
                _y,
                _y_prob,
                sample_weight=sample_weights,
                # For multiclass classification where a one-vs-rest precision-recall curve is
                # produced for each class, the positive label is binarized and should not be
                # included in the plot legend
                pos_label=_pos_label if _pos_label == pos_label else None,
            )
            # NB: We return average precision score (AP) instead of AUC because AP is more
            # appropriate for summarizing a precision-recall curve
            ap = sk_metrics.average_precision_score(
                y_true=_y, y_score=_y_prob, pos_label=_pos_label, sample_weight=sample_weights
            )
            return recall, precision, f"AP={ap:.3f}", ap

        xlabel = "Recall"
        ylabel = "Precision"
        title = "Precision recall curve"
        if pos_label:
            xlabel = f"Recall (Positive label: {pos_label})"
            ylabel = f"Precision (Positive label: {pos_label})"
    else:
        assert False, "illegal curve type"

    if is_binomial:
        x_data, y_data, line_label, auc = gen_line_x_y_label_auc(y, y_probs, pos_label)
        data_series = [(line_label, x_data, y_data)]
    else:
        curve_list = []
        for positive_class_index, positive_class in enumerate(labels):
            y_bin, _, y_prob_bin = _get_binary_sum_up_label_pred_prob(
                positive_class_index, positive_class, y, labels, y_probs
            )

            x_data, y_data, line_label, auc = gen_line_x_y_label_auc(
                y_bin, y_prob_bin, _pos_label=1
            )
            curve_list.append((positive_class, x_data, y_data, line_label, auc))

        data_series = [
            (f"label={positive_class},{line_label}", x_data, y_data)
            for positive_class, x_data, y_data, line_label, _ in curve_list
        ]
        auc = [auc for _, _, _, _, auc in curve_list]

    def _do_plot(**kwargs):
        from matplotlib import pyplot

        _, ax = plot_lines(**kwargs)
        dash_line_args = {
            "color": "gray",
            "alpha": 0.3,
            "drawstyle": "default",
            "linestyle": "dashed",
        }
        if curve_type == "pr":
            ax.plot([0, 1], [1, 0], **dash_line_args)
        elif curve_type == "roc":
            ax.plot([0, 1], [0, 1], **dash_line_args)

        if is_binomial:
            ax.legend(loc="best")
        else:
            ax.legend(loc="center left", bbox_to_anchor=(1, 0.5))
            pyplot.subplots_adjust(right=0.6, bottom=0.25)

    return _Curve(
        plot_fn=_do_plot,
        plot_fn_args={
            "data_series": data_series,
            "xlabel": xlabel,
            "ylabel": ylabel,
            "line_kwargs": {"drawstyle": "steps-post", "linewidth": 1},
            "title": title,
        },
        auc=auc,
    )



import logging

from mlflow.langchain.constant import FLAVOR_NAME
from mlflow.utils.annotations import experimental
from mlflow.utils.autologging_utils import autologging_integration
from mlflow.utils.autologging_utils.config import AutoLoggingConfig
from mlflow.utils.autologging_utils.safety import safe_patch

logger = logging.getLogger(__name__)


@experimental(version="2.10.0")
@autologging_integration(FLAVOR_NAME)
def autolog(
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    log_traces=True,
):
    """
    Enables (or disables) and configures autologging from Langchain to MLflow.

    Args:
        disable: If ``True``, disables the Langchain autologging integration. If ``False``,
            enables the Langchain autologging integration.
        exclusive: If ``True``, autologged content is not logged to user-created fluent runs.
            If ``False``, autologged content is logged to the active fluent run,
            which may be user-created.
        disable_for_unsupported_versions: If ``True``, disable autologging for versions of
            langchain that have not been tested against this version of the MLflow
            client or are incompatible.
        silent: If ``True``, suppress all event logs and warnings from MLflow during Langchain
            autologging. If ``False``, show all events and warnings during Langchain
            autologging.
        log_traces: If ``True``, traces are logged for Langchain models by using
            MlflowLangchainTracer as a callback during inference. If ``False``, no traces are
            collected during inference. Default to ``True``.
    """
    try:
        from langchain_core.callbacks import BaseCallbackManager

        safe_patch(
            FLAVOR_NAME,
            BaseCallbackManager,
            "__init__",
            _patched_callback_manager_init,
        )
    except Exception as e:
        logger.warning(f"Failed to enable tracing for LangChain. Error: {e}")

    # Special handlings for edge cases.
    try:
        from langchain_core.callbacks import BaseCallbackManager
        from langchain_core.runnables import RunnableSequence

        safe_patch(
            FLAVOR_NAME,
            RunnableSequence,
            "batch",
            _patched_runnable_sequence_batch,
        )

        safe_patch(
            FLAVOR_NAME,
            BaseCallbackManager,
            "merge",
            _patched_callback_manager_merge,
        )
    except Exception:
        logger.debug("Failed to patch RunnableSequence or BaseCallbackManager.", exc_info=True)


def _patched_callback_manager_init(original, self, *args, **kwargs):
    from mlflow.langchain.langchain_tracer import MlflowLangchainTracer

    original(self, *args, **kwargs)

    if not AutoLoggingConfig.init(FLAVOR_NAME).log_traces:
        return

    for handler in self.inheritable_handlers:
        if isinstance(handler, MlflowLangchainTracer):
            return

    _handler = MlflowLangchainTracer()
    self.add_handler(_handler, inherit=True)


def _patched_callback_manager_merge(original, self, *args, **kwargs):
    """
    Patch BaseCallbackManager.merge to avoid a duplicated callback issue.

    In the above patched __init__, we check `inheritable_handlers` to see if the MLflow tracer
    is already propagated. This works when the `inheritable_handlers` is specified as constructor
    arguments. However, in the `merge` method, LangChain does not use constructor but set
    callbacks via the setter method. This causes duplicated callbacks injection.
    https://github.com/langchain-ai/langchain/blob/d9a069c414a321e7a3f3638a32ecf8a37ec2d188/libs/core/langchain_core/callbacks/base.py#L962-L982
    """
    from mlflow.langchain.langchain_tracer import MlflowLangchainTracer

    # Get the MLflow callback inherited from parent
    inherited = self.inheritable_handlers + args[0].inheritable_handlers
    inherited_mlflow_cb = next(
        (cb for cb in inherited if isinstance(cb, MlflowLangchainTracer)), None
    )

    if not inherited_mlflow_cb:
        return original(self, *args, **kwargs)

    merged = original(self, *args, **kwargs)
    # If a new MLflow callback is generated inside __init__, remove it
    duplicate_mlflow_cbs = [
        cb
        for cb in merged.inheritable_handlers
        if isinstance(cb, MlflowLangchainTracer) and cb != inherited_mlflow_cb
    ]
    for cb in duplicate_mlflow_cbs:
        merged.remove_handler(cb)

    return merged


def _patched_runnable_sequence_batch(original, self, *args, **kwargs):
    """
    Patch to terminate span context attachment during batch execution.

    RunnableSequence's batch() methods are implemented in a peculiar way
    that iterates on steps->items sequentially within the same thread. For example, if a
    sequence has 2 steps and the batch size is 3, the execution flow will be:
      - Step 1 for item 1
      - Step 1 for item 2
      - Step 1 for item 3
      - Step 2 for item 1
      - Step 2 for item 2
      - Step 2 for item 3
    Due to this behavior, we cannot attach the span to the context for this particular
    API, otherwise spans for different inputs will be mixed up.
    """
    from mlflow.langchain.langchain_tracer import _should_attach_span_to_context

    original_state = _should_attach_span_to_context.get()
    _should_attach_span_to_context.set(False)
    try:
        return original(self, *args, **kwargs)
    finally:
        _should_attach_span_to_context.set(original_state)



"""
Built-in Plugins for Genesis-Flow

This module contains all built-in plugins that ship with Genesis-Flow.
These plugins provide integration with essential ML frameworks.
"""

from mlflow.plugins.builtin.pytorch_plugin import PyTorchPlugin
from mlflow.plugins.builtin.sklearn_plugin import SklearnPlugin
from mlflow.plugins.builtin.transformers_plugin import TransformersPlugin

# Registry of all built-in plugins
BUILTIN_PLUGINS = {
    "pytorch": PyTorchPlugin,
    "sklearn": SklearnPlugin,
    "transformers": TransformersPlugin,
}

__all__ = [
    "BUILTIN_PLUGINS",
    "PyTorchPlugin",
    "SklearnPlugin", 
    "TransformersPlugin",
]


"""
PyTorch Plugin for Genesis-Flow

Provides PyTorch integration as a modular plugin that can be enabled/disabled.
"""

import logging
from typing import Dict, Callable

from mlflow.plugins.base import FrameworkPlugin, PluginMetadata, PluginType

logger = logging.getLogger(__name__)

class PyTorchPlugin(FrameworkPlugin):
    """
    PyTorch framework plugin for Genesis-Flow.
    
    Provides PyTorch model logging, autologging, and deployment capabilities.
    """
    
    @classmethod
    def get_default_metadata(cls) -> PluginMetadata:
        """Get default metadata for PyTorch plugin."""
        return PluginMetadata(
            name="pytorch",
            version="1.0.0",
            description="PyTorch deep learning framework integration",
            author="Genesis-Flow Team",
            plugin_type=PluginType.FRAMEWORK,
            dependencies=["torch"],
            optional_dependencies=["torchvision", "torchaudio"],
            min_genesis_flow_version="1.0.0",
            homepage="https://pytorch.org",
            documentation="https://mlflow.org/docs/latest/models.html#pytorch-pytorch",
            license="Apache 2.0",
            tags=["deep-learning", "neural-networks", "pytorch"],
        )
    
    def get_module_path(self) -> str:
        """Get the module path for PyTorch integration."""
        return "mlflow.pytorch"
    
    def get_autolog_functions(self) -> Dict[str, Callable]:
        """Get PyTorch autologging functions."""
        try:
            from mlflow.pytorch import autolog
            return {
                "autolog": autolog,
            }
        except ImportError:
            logger.warning("PyTorch autolog not available")
            return {}
    
    def get_save_functions(self) -> Dict[str, Callable]:
        """Get PyTorch model saving functions."""
        try:
            from mlflow.pytorch import log_model, save_model
            return {
                "log_model": log_model,
                "save_model": save_model,
            }
        except ImportError:
            logger.warning("PyTorch save functions not available")
            return {}
    
    def get_load_functions(self) -> Dict[str, Callable]:
        """Get PyTorch model loading functions."""
        try:
            from mlflow.pytorch import load_model
            return {
                "load_model": load_model,
            }
        except ImportError:
            logger.warning("PyTorch load functions not available")
            return {}
    
    def enable(self) -> bool:
        """Enable PyTorch plugin with enhanced functionality."""
        try:
            # Call parent enable
            if not super().enable():
                return False
            
            # Additional PyTorch-specific setup
            self._setup_pytorch_environment()
            
            self._logger.info("PyTorch plugin enabled with full functionality")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to enable PyTorch plugin: {e}")
            return False
    
    def _setup_pytorch_environment(self):
        """Setup PyTorch-specific environment and configurations."""
        try:
            import torch
            
            # Log PyTorch version and CUDA availability
            self._logger.info(f"PyTorch version: {torch.__version__}")
            self._logger.info(f"CUDA available: {torch.cuda.is_available()}")
            
            if torch.cuda.is_available():
                self._logger.info(f"CUDA version: {torch.version.cuda}")
                self._logger.info(f"GPU count: {torch.cuda.device_count()}")
            
            # Setup hook for automatic model artifact logging
            self.register_hook("model_save", self._on_model_save)
            
        except Exception as e:
            self._logger.warning(f"PyTorch environment setup warning: {e}")
    
    def _on_model_save(self, model, *args, **kwargs):
        """Hook called when a model is saved."""
        try:
            import torch
            
            if isinstance(model, torch.nn.Module):
                self._logger.debug("PyTorch model save detected")
                # Additional logging or processing can be added here
                
        except Exception as e:
            self._logger.warning(f"PyTorch model save hook error: {e}")
    
    def check_dependencies(self) -> bool:
        """Enhanced dependency checking for PyTorch."""
        # Check basic dependencies first
        if not super().check_dependencies():
            return False
        
        try:
            import torch
            
            # Check minimum PyTorch version
            from packaging.version import Version
            min_pytorch_version = "1.9.0"
            
            if Version(torch.__version__) < Version(min_pytorch_version):
                self._logger.error(f"PyTorch {min_pytorch_version}+ required, found {torch.__version__}")
                return False
            
            self._logger.debug(f"PyTorch {torch.__version__} dependency satisfied")
            return True
            
        except ImportError:
            self._logger.error("PyTorch not installed")
            return False
        except Exception as e:
            self._logger.error(f"PyTorch dependency check failed: {e}")
            return False


import os
from typing import Any, Optional, Union

import yaml

from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE

__mlflow_model_config__ = None


class ModelConfig:
    """
    ModelConfig used in code to read a YAML configuration file or a dictionary.

    Args:
        development_config: Path to the YAML configuration file or a dictionary containing the
                        configuration. If the configuration is not provided, an error is raised

    .. code-block:: python
        :caption: Example usage in model code

        from mlflow.models import ModelConfig

        # Load the configuration from a dictionary
        config = ModelConfig(development_config={"key1": "value1"})
        print(config.get("key1"))


    .. code-block:: yaml
        :caption: yaml file for model configuration

        key1: value1
        another_key:
            - value2
            - value3

    .. code-block:: python
        :caption: Example yaml usage in model code

        from mlflow.models import ModelConfig

        # Load the configuration from a file
        config = ModelConfig(development_config="config.yaml")
        print(config.get("key1"))


    When invoking the ModelConfig locally in a model file, development_config can be passed in
    which would be used as configuration for the model.


    .. code-block:: python
        :caption: Example to use ModelConfig when logging model as code: agent.py

        import mlflow
        from mlflow.models import ModelConfig

        config = ModelConfig(development_config={"key1": "value1"})


        class TestModel(mlflow.pyfunc.PythonModel):
            def predict(self, context, model_input, params=None):
                return config.get("key1")


        mlflow.models.set_model(TestModel())


    But this development_config configuration file will be overridden when logging a model.
    When no model_config is passed in while logging the model, an error will be raised when
    trying to load the model using ModelConfig.
    Note: development_config is not used when logging the model.


    .. code-block:: python
        :caption: Example to use agent.py to log the model: deploy.py

        model_config = {"key1": "value2"}
        with mlflow.start_run():
            model_info = mlflow.pyfunc.log_model(
                name="model", python_model="agent.py", model_config=model_config
            )

        loaded_model = mlflow.pyfunc.load_model(model_info.model_uri)

        # This will print "value2" as the model_config passed in while logging the model
        print(loaded_model.predict(None))
    """

    def __init__(self, *, development_config: Optional[Union[str, dict[str, Any]]] = None):
        config = globals().get("__mlflow_model_config__", None)
        # Here mlflow_model_config have 3 states:
        # 1. None, this means if the mlflow_model_config is None, use development_config if
        # available
        # 2. "", Empty string, this means the users explicitly didn't set the model config
        # while logging the model so if ModelConfig is used, it should throw an error
        # 3. A valid path, this means the users have set the model config while logging the
        # model so use that path
        if config is not None:
            self.config = config
        else:
            self.config = development_config

        if not self.config:
            raise FileNotFoundError(
                "Config file is not provided which is needed to load the model. "
                "Please provide a valid path."
            )

        if not isinstance(self.config, dict) and not os.path.isfile(self.config):
            raise FileNotFoundError(f"Config file '{self.config}' not found.")

    def _read_config(self):
        """Reads the YAML configuration file and returns its contents.

        Raises:
            FileNotFoundError: If the configuration file does not exist.
            yaml.YAMLError: If there is an error parsing the YAML content.

        Returns:
            dict or None: The content of the YAML file as a dictionary, or None if the
            config path is not set.
        """
        if isinstance(self.config, dict):
            return self.config

        with open(self.config) as file:
            try:
                return yaml.safe_load(file)
            except yaml.YAMLError as e:
                raise MlflowException(
                    f"Error parsing YAML file: {e}", error_code=INVALID_PARAMETER_VALUE
                )

    def to_dict(self):
        """Returns the configuration as a dictionary."""
        return self._read_config()

    def get(self, key):
        """Gets the value of a top-level parameter in the configuration."""
        config_data = self._read_config()

        if config_data and key in config_data:
            return config_data[key]
        else:
            raise KeyError(f"Key '{key}' not found in configuration: {config_data}.")


def _set_model_config(model_config):
    globals()["__mlflow_model_config__"] = model_config



import functools
import logging
from typing import Optional

import numpy as np
from packaging.version import Version
from sklearn.pipeline import Pipeline as sk_Pipeline

import mlflow
from mlflow import MlflowException
from mlflow.models.evaluation.base import EvaluationMetric, EvaluationResult, _ModelType
from mlflow.models.evaluation.default_evaluator import (
    BuiltInEvaluator,
    _extract_predict_fn,
    _extract_raw_model,
    _get_dataframe_with_renamed_columns,
)
from mlflow.models.evaluation.evaluators.classifier import (
    _is_continuous,
    _suppress_class_imbalance_errors,
)
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.pyfunc import _ServedPyFuncModel

_logger = logging.getLogger(__name__)


_SUPPORTED_SHAP_ALGORITHMS = ("exact", "permutation", "partition", "kernel")
_DEFAULT_SAMPLE_ROWS_FOR_SHAP = 2000


def _shap_predict_fn(x, predict_fn, feature_names):
    return predict_fn(_get_dataframe_with_renamed_columns(x, feature_names))


class ShapEvaluator(BuiltInEvaluator):
    """
    A built-in evaluator to get SHAP explainability insights for classifier and regressor models.

    This evaluator often run with the main evaluator for the model like ClassifierEvaluator.
    """

    name = "shap"

    @classmethod
    def can_evaluate(cls, *, model_type, evaluator_config, **kwargs):
        return model_type in (_ModelType.CLASSIFIER, _ModelType.REGRESSOR) and evaluator_config.get(
            "log_model_explainability", True
        )

    def _evaluate(
        self,
        model: Optional["mlflow.pyfunc.PyFuncModel"],
        extra_metrics: list[EvaluationMetric],
        custom_artifacts=None,
        **kwargs,
    ) -> Optional[EvaluationResult]:
        if isinstance(model, _ServedPyFuncModel):
            _logger.warning(
                "Skipping model explainability because a model server is used for environment "
                "restoration."
            )
            return

        model_loader_module, raw_model = _extract_raw_model(model)
        if model_loader_module == "mlflow.spark":
            # TODO: Shap explainer need to manipulate on each feature values,
            #  but spark model input dataframe contains Vector type feature column
            #  which shap explainer does not support.
            #  To support this, we need expand the Vector type feature column into
            #  multiple scalar feature columns and pass it to shap explainer.
            _logger.warning(
                "Logging model explainability insights is not currently supported for PySpark "
                "models."
            )
            return

        self.y_true = self.dataset.labels_data
        self.label_list = self.evaluator_config.get("label_list")
        self.pos_label = self.evaluator_config.get("pos_label")

        if not (np.issubdtype(self.y_true.dtype, np.number) or self.y_true.dtype == np.bool_):
            # Note: python bool type inherits number type but np.bool_ does not inherit np.number.
            _logger.warning(
                "Skip logging model explainability insights because it requires all label "
                "values to be numeric or boolean."
            )
            return

        algorithm = self.evaluator_config.get("explainability_algorithm", None)
        if algorithm is not None and algorithm not in _SUPPORTED_SHAP_ALGORITHMS:
            raise MlflowException(
                message=f"Specified explainer algorithm {algorithm} is unsupported. Currently only "
                f"support {','.join(_SUPPORTED_SHAP_ALGORITHMS)} algorithms.",
                error_code=INVALID_PARAMETER_VALUE,
            )

        if algorithm != "kernel":
            feature_dtypes = list(self.X.get_original().dtypes)
            for feature_dtype in feature_dtypes:
                if not np.issubdtype(feature_dtype, np.number):
                    _logger.warning(
                        "Skip logging model explainability insights because the shap explainer "
                        f"{algorithm} requires all feature values to be numeric, and each feature "
                        "column must only contain scalar values."
                    )
                    return

        try:
            import shap
            from matplotlib import pyplot
        except ImportError:
            _logger.warning(
                "SHAP or matplotlib package is not installed, so model explainability insights "
                "will not be logged."
            )
            return

        if Version(shap.__version__) < Version("0.40"):
            _logger.warning(
                "Shap package version is lower than 0.40, Skip log model explainability."
            )
            return

        sample_rows = self.evaluator_config.get(
            "explainability_nsamples", _DEFAULT_SAMPLE_ROWS_FOR_SHAP
        )

        X_df = self.X.copy_to_avoid_mutation()

        sampled_X = shap.sample(X_df, sample_rows, random_state=0)

        mode_or_mean_dict = _compute_df_mode_or_mean(X_df)
        sampled_X = sampled_X.fillna(mode_or_mean_dict)

        # shap explainer might call provided `predict_fn` with a `numpy.ndarray` type
        # argument, this might break some model inference, so convert the argument into
        # a pandas dataframe.
        # The `shap_predict_fn` calls model's predict function, we need to restore the input
        # dataframe with original column names, because some model prediction routine uses
        # the column name.

        predict_fn = _extract_predict_fn(model)
        shap_predict_fn = functools.partial(
            _shap_predict_fn, predict_fn=predict_fn, feature_names=self.dataset.feature_names
        )

        if self.label_list is None:
            # If label list is not specified, infer label list from model output.
            # We need to copy the input data as the model might mutate the input data.
            y_pred = predict_fn(X_df.copy()) if predict_fn else self.dataset.predictions_data
            self.label_list = np.unique(np.concatenate([self.y_true, y_pred]))

        try:
            if algorithm:
                if algorithm == "kernel":
                    # We need to lazily import shap, so lazily import `_PatchedKernelExplainer`
                    from mlflow.models.evaluation._shap_patch import _PatchedKernelExplainer

                    kernel_link = self.evaluator_config.get(
                        "explainability_kernel_link", "identity"
                    )
                    if kernel_link not in ["identity", "logit"]:
                        raise ValueError(
                            "explainability_kernel_link config can only be set to 'identity' or "
                            f"'logit', but got '{kernel_link}'."
                        )
                    background_X = shap.sample(X_df, sample_rows, random_state=3)
                    background_X = background_X.fillna(mode_or_mean_dict)

                    explainer = _PatchedKernelExplainer(
                        shap_predict_fn, background_X, link=kernel_link
                    )
                else:
                    explainer = shap.Explainer(
                        shap_predict_fn,
                        sampled_X,
                        feature_names=self.dataset.feature_names,
                        algorithm=algorithm,
                    )
            else:
                if (
                    raw_model
                    and not len(self.label_list) > 2
                    and not isinstance(raw_model, sk_Pipeline)
                ):
                    # For mulitnomial classifier, shap.Explainer may choose Tree/Linear explainer
                    # for raw model, this case shap plot doesn't support it well, so exclude the
                    # multinomial_classifier case here.
                    explainer = shap.Explainer(
                        raw_model, sampled_X, feature_names=self.dataset.feature_names
                    )
                else:
                    # fallback to default explainer
                    explainer = shap.Explainer(
                        shap_predict_fn, sampled_X, feature_names=self.dataset.feature_names
                    )

            _logger.info(f"Shap explainer {explainer.__class__.__name__} is used.")

            if algorithm == "kernel":
                shap_values = shap.Explanation(
                    explainer.shap_values(sampled_X), feature_names=self.dataset.feature_names
                )
            else:
                shap_values = explainer(sampled_X)
        except Exception as e:
            # Shap evaluation might fail on some edge cases, e.g., unsupported input data values
            # or unsupported model on specific shap explainer. Catch exception to prevent it
            # breaking the whole `evaluate` function.

            if not self.evaluator_config.get("ignore_exceptions", True):
                raise e

            _logger.warning(
                f"Shap evaluation failed. Reason: {e!r}. "
                "Set logging level to DEBUG to see the full traceback."
            )
            _logger.debug("", exc_info=True)
            return

        if self.evaluator_config.get("log_explainer", False):
            try:
                mlflow.shap.log_explainer(explainer, name="explainer")
            except Exception as e:
                # TODO: The explainer saver is buggy, if `get_underlying_model_flavor` return
                #   "unknown", then fallback to shap explainer saver, and shap explainer will call
                #   `model.save` for sklearn model, there is no `.save` method, so error will
                #   happen.
                _logger.warning(
                    f"Logging explainer failed. Reason: {e!r}. "
                    "Set logging level to DEBUG to see the full traceback."
                )
                _logger.debug("", exc_info=True)

        def _adjust_color_bar():
            pyplot.gcf().axes[-1].set_aspect("auto")
            pyplot.gcf().axes[-1].set_box_aspect(50)

        def _adjust_axis_tick():
            pyplot.xticks(fontsize=10)
            pyplot.yticks(fontsize=10)

        def plot_beeswarm():
            shap.plots.beeswarm(shap_values, show=False, color_bar=True)
            _adjust_color_bar()
            _adjust_axis_tick()

        with _suppress_class_imbalance_errors(ValueError, log_warning=False):
            self._log_image_artifact(
                plot_beeswarm,
                "shap_beeswarm_plot",
            )

        def plot_summary():
            shap.summary_plot(shap_values, show=False, color_bar=True)
            _adjust_color_bar()
            _adjust_axis_tick()

        with _suppress_class_imbalance_errors(TypeError, log_warning=False):
            self._log_image_artifact(
                plot_summary,
                "shap_summary_plot",
            )

        def plot_feature_importance():
            shap.plots.bar(shap_values, show=False)
            _adjust_axis_tick()

        with _suppress_class_imbalance_errors(IndexError, log_warning=False):
            self._log_image_artifact(
                plot_feature_importance,
                "shap_feature_importance_plot",
            )

        return EvaluationResult(
            metrics=self.aggregate_metrics,
            artifacts=self.artifacts,
            run_id=self.run_id,
        )


def _compute_df_mode_or_mean(df):
    """
    Compute mean (for continuous columns) and compute mode (for other columns) for the
    input dataframe, return a dict, key is column name, value is the corresponding mode or
    mean value, this function calls `_is_continuous` to determine whether the
    column is continuous column.
    """
    continuous_cols = [c for c in df.columns if _is_continuous(df[c])]
    df_cont = df[continuous_cols]
    df_non_cont = df.drop(continuous_cols, axis=1)

    means = {} if df_cont.empty else df_cont.mean().to_dict()
    modes = {} if df_non_cont.empty else df_non_cont.mode().loc[0].to_dict()
    return {**means, **modes}



import importlib.metadata
import json
from dataclasses import asdict, is_dataclass
from typing import TYPE_CHECKING, Any, Optional, Union

from packaging.version import Version

if TYPE_CHECKING:
    import dspy

from mlflow.exceptions import INVALID_PARAMETER_VALUE, MlflowException
from mlflow.protos.databricks_pb2 import (
    INVALID_PARAMETER_VALUE,
)
from mlflow.pyfunc import PythonModel
from mlflow.types.schema import DataType, Schema

_INVALID_SIZE_MESSAGE = (
    "Dspy model doesn't support batch inference or empty input. Please provide a single input."
)


class DspyModelWrapper(PythonModel):
    """MLflow PyFunc wrapper class for Dspy models.

    This wrapper serves two purposes:
        - It stores the Dspy model along with dspy global settings, which are required for seamless
            saving and loading.
        - It provides a `predict` method so that it can be loaded as an MLflow pyfunc, which is
            used at serving time.
    """

    def __init__(
        self,
        model: "dspy.Module",
        dspy_settings: dict[str, Any],
        model_config: Optional[dict[str, Any]] = None,
    ):
        self.model = model
        self.dspy_settings = dspy_settings
        self.model_config = model_config or {}
        self.output_schema: Optional[Schema] = None

    def predict(self, inputs: Any, params: Optional[dict[str, Any]] = None):
        import dspy

        converted_inputs = self._get_model_input(inputs)

        with dspy.context(**self.dspy_settings):
            if isinstance(converted_inputs, dict):
                # We pass a dict as keyword args and don't allow DSPy models
                # to receive a single dict.
                return self.model(**converted_inputs).toDict()
            else:
                return self.model(converted_inputs).toDict()

    def predict_stream(self, inputs: Any, params=None):
        import dspy

        converted_inputs = self._get_model_input(inputs)

        self._validate_streaming()

        stream_listeners = [
            dspy.streaming.StreamListener(signature_field_name=spec.name)
            for spec in self.output_schema
        ]
        stream_model = dspy.streamify(
            self.model,
            stream_listeners=stream_listeners,
            async_streaming=False,
            include_final_prediction_in_output_stream=False,
        )

        if isinstance(converted_inputs, dict):
            outputs = stream_model(**converted_inputs)
        else:
            outputs = stream_model(converted_inputs)

        with dspy.context(**self.dspy_settings):
            for output in outputs:
                if is_dataclass(output):
                    yield asdict(output)
                elif isinstance(output, dspy.Prediction):
                    yield output.toDict()
                else:
                    yield output

    def _get_model_input(self, inputs: Any) -> Union[str, dict[str, Any]]:
        """Convert the PythonModel input into the DSPy program input

        Examples of expected conversions:
        - str -> str
        - dict -> dict
        - np.ndarray with one element -> single element
        - pd.DataFrame with one row and string column -> single row dict
        - pd.DataFrame with one row and non-string column -> single element
        - list -> raises an exception
        - np.ndarray with more than one element -> raises an exception
        - pd.DataFrame with more than one row -> raises an exception
        """
        import numpy as np
        import pandas as pd

        supported_input_types = (np.ndarray, pd.DataFrame, str, dict)
        if not isinstance(inputs, supported_input_types):
            raise MlflowException(
                f"`inputs` must be one of: {[x.__name__ for x in supported_input_types]}, but "
                f"received type: {type(inputs)}.",
                INVALID_PARAMETER_VALUE,
            )
        if isinstance(inputs, pd.DataFrame):
            if len(inputs) != 1:
                raise MlflowException(
                    _INVALID_SIZE_MESSAGE,
                    INVALID_PARAMETER_VALUE,
                )
            if all(isinstance(col, str) for col in inputs.columns):
                inputs = inputs.to_dict(orient="records")[0]
            else:
                inputs = inputs.values[0]
        if isinstance(inputs, np.ndarray):
            if len(inputs) != 1:
                raise MlflowException(
                    _INVALID_SIZE_MESSAGE,
                    INVALID_PARAMETER_VALUE,
                )
            inputs = inputs[0]

        return inputs

    def _validate_streaming(
        self,
    ):
        if Version(importlib.metadata.version("dspy")) <= Version("2.6.23"):
            raise MlflowException(
                "Streaming API is only supported in dspy 2.6.24 or later. "
                "Please upgrade your dspy version."
            )

        if self.output_schema is None:
            raise MlflowException(
                "Output schema of the DSPy model is not set. Please log your DSPy "
                "model with `signature` or `input_example` to use streaming API."
            )

        if any(spec.type != DataType.string for spec in self.output_schema):
            raise MlflowException(
                f"All output fields must be string to use streaming API. Got {self.output_schema}."
            )


class DspyChatModelWrapper(DspyModelWrapper):
    """MLflow PyFunc wrapper class for Dspy chat models."""

    def predict(self, inputs: Any, params: Optional[dict[str, Any]] = None):
        import dspy

        converted_inputs = self._get_model_input(inputs)

        # `dspy.settings` cannot be shared across threads, so we are setting the context at every
        # predict call.
        with dspy.context(**self.dspy_settings):
            outputs = self.model(converted_inputs)

        choices = []
        if isinstance(outputs, str):
            choices.append(self._construct_chat_message("assistant", outputs))
        elif isinstance(outputs, dict):
            role = outputs.get("role", "assistant")
            choices.append(self._construct_chat_message(role, json.dumps(outputs)))
        elif isinstance(outputs, dspy.Prediction):
            choices.append(self._construct_chat_message("assistant", json.dumps(outputs.toDict())))
        elif isinstance(outputs, list):
            for output in outputs:
                if isinstance(output, dict):
                    role = output.get("role", "assistant")
                    choices.append(self._construct_chat_message(role, json.dumps(outputs)))
                elif isinstance(output, dspy.Prediction):
                    role = output.get("role", "assistant")
                    choices.append(self._construct_chat_message(role, json.dumps(outputs.toDict())))
                else:
                    raise MlflowException(
                        f"Unsupported output type: {type(output)}. To log a DSPy model with task "
                        "'llm/v1/chat', the DSPy model must return a dict, a dspy.Prediction, or a "
                        "list of dicts or dspy.Prediction.",
                        INVALID_PARAMETER_VALUE,
                    )
        else:
            raise MlflowException(
                f"Unsupported output type: {type(outputs)}. To log a DSPy model with task "
                "'llm/v1/chat', the DSPy model must return a dict, a dspy.Prediction, or a list of "
                "dicts or dspy.Prediction.",
                INVALID_PARAMETER_VALUE,
            )

        return {"choices": choices}

    def predict_stream(self, inputs: Any, params=None):
        raise NotImplementedError(
            "Streaming is not supported for DSPy model with task 'llm/v1/chat'."
        )

    def _get_model_input(self, inputs: Any) -> Union[str, list[dict[str, Any]]]:
        import pandas as pd

        if isinstance(inputs, dict):
            return inputs["messages"]
        if isinstance(inputs, pd.DataFrame):
            return inputs.messages[0]

        raise MlflowException(
            f"Unsupported input type: {type(inputs)}. To log a DSPy model with task "
            "'llm/v1/chat', the input must be a dict or a pandas DataFrame.",
            INVALID_PARAMETER_VALUE,
        )

    def _construct_chat_message(self, role: str, content: str) -> dict[str, Any]:
        return {
            "index": 0,
            "message": {
                "role": role,
                "content": content,
            },
            "finish_reason": "stop",
        }



import importlib.metadata
import logging
from contextlib import contextmanager
from threading import Thread

from packaging.version import Version

from mlflow.utils.annotations import experimental
from mlflow.utils.autologging_utils import (
    autologging_integration,
    safe_patch,
)
from mlflow.utils.databricks_utils import is_in_databricks_runtime

FLAVOR_NAME = "litellm"

_logger = logging.getLogger(__name__)


@experimental(version="2.19.0")
def autolog(
    log_traces: bool = True,
    disable: bool = False,
    silent: bool = False,
):
    """
    Enables (or disables) and configures autologging from LiteLLM to MLflow. Currently, MLflow
    only supports autologging for tracing.

    Args:
        log_traces: If ``True``, traces are logged for LiteLLM calls. If ``False``,
            no traces are collected during inference. Default to ``True``.
        disable: If ``True``, disables the LiteLLM autologging integration. If ``False``,
            enables the LiteLLM autologging integration.
        silent: If ``True``, suppress all event logs and warnings from MLflow during LiteLLM
            autologging. If ``False``, show all events and warnings.
    """
    import litellm

    # This needs to be called before doing any safe-patching (otherwise safe-patch will be no-op).
    # TODO: since this implementation is inconsistent, explore a universal way to solve the issue.
    _autolog(log_traces=log_traces, disable=disable, silent=silent)

    try:
        from litellm.integrations.mlflow import MlflowLogger
    except ImportError:
        _logger.warning(
            "MLflow LiteLLM integration is not supported for the installed LiteLLM version. "
            "Please upgrade to a newer version to enable MLflow LiteLLM autologging."
        )
        return

    if log_traces and not disable:
        litellm.success_callback = _append_mlflow_callbacks(litellm.success_callback)
        litellm.failure_callback = _append_mlflow_callbacks(litellm.failure_callback)

        # Workaround for https://github.com/BerriAI/litellm/issues/8013
        # TODO: Add upper bound version check when the issue is fixed.
        if Version(importlib.metadata.version("litellm")) >= Version("1.59.4"):
            litellm.failure_callback = [
                cb if cb != "mlflow" else MlflowLogger() for cb in litellm.failure_callback
            ]

        if is_in_databricks_runtime():
            # Patch main APIs e.g. completion to inject custom handling for threading.
            # By default, those API will start a new thread when calling log_success_event()
            # handler of the logging callbacks and never wait for it to finish. This is
            # problematic in Databricks notebook, because the inline trace UI display
            # assumes that the trace is generated synchronously. If the trace is generated
            # asynchronously, it will be displayed in different later cells.
            # To workaround this issue, we monkey-patch these APIs to wait for the logging
            # threads to finish before returning the result.
            # This is not required for OSS environment where we don't show inline trace UI.
            for func in [
                "completion",
                "embedding",
                "text_completion",
                "image_generation",
                "transcription",
                "speech",
            ]:
                _patch_threading_in_function(litellm, func)

            # For streaming case, we need to patch the iterator because traces are generated
            # when consuming the generator, not when calling the main APIs.
            _patch_threading_in_function(litellm.utils.CustomStreamWrapper, "__next__")

            # NB: We don't need to patch async function because Databricks notebook waits
            # for the async task to finish before finishing the cell.
    else:
        litellm.success_callback = _remove_mlflow_callbacks(litellm.success_callback)
        litellm.failure_callback = _remove_mlflow_callbacks(litellm.failure_callback)
        # Callback also needs to be removed from 'callbacks' as litellm adds
        # success/failure callbacks to there as well.
        litellm.callbacks = _remove_mlflow_callbacks(litellm.callbacks)


# This is required by mlflow.autolog()
autolog.integration_name = FLAVOR_NAME


# NB: The @autologging_integration annotation must be applied here, and the callback injection
# needs to happen outside the annotated function. This is because the annotated function is NOT
# executed when disable=True is passed. This prevents us from removing our callback and patching
# when autologging is turned off.
@autologging_integration(FLAVOR_NAME)
def _autolog(
    log_traces: bool,
    disable: bool = False,
    silent: bool = False,
):
    pass


def _patch_threading_in_function(target, function_name: str):
    """
    Apply the threading patch to a synchronous function.

    We capture the threads started by the function using the _patch_thread_start context manager,
    then join them to ensure they are finished before the notebook cell finishes executing.
    """

    def _patch_fn(original, *args, **kwargs):
        with _patch_thread_start() as logging_threads:
            result = original(*args, **kwargs)
        for thread in logging_threads:
            thread.join()
        return result

    safe_patch(FLAVOR_NAME, target, function_name, _patch_fn)


@contextmanager
def _patch_thread_start():
    """
    A context manager to collect threads started for logging handlers.
    This is done by monkey-patching the start() method of threading.Thread.
    Note that failure handlers are executed synchronously, so we don't need to patch them.
    """
    original = Thread.start
    logging_threads = []

    def patched_thread(self, *args, **kwargs):
        target = getattr(self, "_target", None)
        # success_handler is for normal request, and run_success_... is for streaming
        # - https://github.com/BerriAI/litellm/blob/4f8a3fd4cfc20cf43b38379928b41c2691c85d36/litellm/utils.py#L946
        # - https://github.com/BerriAI/litellm/blob/4f8a3fd4cfc20cf43b38379928b41c2691c85d36/litellm/utils.py#L7526
        if target and target.__name__ in [
            "success_handler",
            "run_success_logging_and_cache_storage",
        ]:
            logging_threads.append(self)
        return original(self, *args, **kwargs)

    Thread.start = patched_thread
    try:
        yield logging_threads
    finally:
        Thread.start = original


def _append_mlflow_callbacks(callbacks):
    from litellm.integrations.mlflow import MlflowLogger

    # MLflow callback can be stored as a string or the actual logger object
    if not any(cb == "mlflow" or isinstance(cb, MlflowLogger) for cb in callbacks):
        return callbacks + ["mlflow"]

    return callbacks


def _remove_mlflow_callbacks(callbacks):
    from litellm.integrations.mlflow import MlflowLogger

    return [cb for cb in callbacks if not (cb == "mlflow" or isinstance(cb, MlflowLogger))]



"""
Scikit-learn Plugin for Genesis-Flow

Provides scikit-learn integration as a modular plugin.
"""

import logging
from typing import Dict, Callable

from mlflow.plugins.base import FrameworkPlugin, PluginMetadata, PluginType

logger = logging.getLogger(__name__)

class SklearnPlugin(FrameworkPlugin):
    """
    Scikit-learn framework plugin for Genesis-Flow.
    
    Provides scikit-learn model logging, autologging, and deployment capabilities.
    """
    
    @classmethod
    def get_default_metadata(cls) -> PluginMetadata:
        """Get default metadata for scikit-learn plugin."""
        return PluginMetadata(
            name="sklearn",
            version="1.0.0",
            description="Scikit-learn machine learning framework integration",
            author="Genesis-Flow Team",
            plugin_type=PluginType.FRAMEWORK,
            dependencies=["scikit-learn"],
            optional_dependencies=["numpy", "scipy", "pandas"],
            min_genesis_flow_version="1.0.0",
            homepage="https://scikit-learn.org",
            documentation="https://mlflow.org/docs/latest/models.html#scikit-learn-sklearn",
            license="Apache 2.0",
            tags=["machine-learning", "scikit-learn", "sklearn"],
        )
    
    def get_module_path(self) -> str:
        """Get the module path for scikit-learn integration."""
        return "mlflow.sklearn"
    
    def get_autolog_functions(self) -> Dict[str, Callable]:
        """Get scikit-learn autologging functions."""
        try:
            from mlflow.sklearn import autolog
            return {
                "autolog": autolog,
            }
        except ImportError:
            logger.warning("Scikit-learn autolog not available")
            return {}
    
    def get_save_functions(self) -> Dict[str, Callable]:
        """Get scikit-learn model saving functions."""
        try:
            from mlflow.sklearn import log_model, save_model
            return {
                "log_model": log_model,
                "save_model": save_model,
            }
        except ImportError:
            logger.warning("Scikit-learn save functions not available")
            return {}
    
    def get_load_functions(self) -> Dict[str, Callable]:
        """Get scikit-learn model loading functions."""
        try:
            from mlflow.sklearn import load_model
            return {
                "load_model": load_model,
            }
        except ImportError:
            logger.warning("Scikit-learn load functions not available")
            return {}
    
    def enable(self) -> bool:
        """Enable scikit-learn plugin with enhanced functionality."""
        try:
            # Call parent enable
            if not super().enable():
                return False
            
            # Additional scikit-learn-specific setup
            self._setup_sklearn_environment()
            
            self._logger.info("Scikit-learn plugin enabled with full functionality")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to enable scikit-learn plugin: {e}")
            return False
    
    def _setup_sklearn_environment(self):
        """Setup scikit-learn-specific environment and configurations."""
        try:
            import sklearn
            
            # Log scikit-learn version
            self._logger.info(f"Scikit-learn version: {sklearn.__version__}")
            
            # Setup hooks for automatic model artifact logging
            self.register_hook("model_save", self._on_model_save)
            self.register_hook("model_fit", self._on_model_fit)
            
        except Exception as e:
            self._logger.warning(f"Scikit-learn environment setup warning: {e}")
    
    def _on_model_save(self, model, *args, **kwargs):
        """Hook called when a model is saved."""
        try:
            import sklearn.base
            
            if isinstance(model, sklearn.base.BaseEstimator):
                self._logger.debug("Scikit-learn model save detected")
                # Additional logging or processing can be added here
                
        except Exception as e:
            self._logger.warning(f"Scikit-learn model save hook error: {e}")
    
    def _on_model_fit(self, model, *args, **kwargs):
        """Hook called when a model is fitted."""
        try:
            import sklearn.base
            
            if isinstance(model, sklearn.base.BaseEstimator):
                self._logger.debug("Scikit-learn model fit detected")
                # Could automatically log model parameters here
                
        except Exception as e:
            self._logger.warning(f"Scikit-learn model fit hook error: {e}")
    
    def check_dependencies(self) -> bool:
        """Enhanced dependency checking for scikit-learn."""
        # Check basic dependencies first
        if not super().check_dependencies():
            return False
        
        try:
            import sklearn
            
            # Check minimum scikit-learn version
            from packaging.version import Version
            min_sklearn_version = "1.0.0"
            
            if Version(sklearn.__version__) < Version(min_sklearn_version):
                self._logger.error(f"Scikit-learn {min_sklearn_version}+ required, found {sklearn.__version__}")
                return False
            
            self._logger.debug(f"Scikit-learn {sklearn.__version__} dependency satisfied")
            return True
            
        except ImportError:
            self._logger.error("Scikit-learn not installed")
            return False
        except Exception as e:
            self._logger.error(f"Scikit-learn dependency check failed: {e}")
            return False